# Copyright (C) 2008-2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


import os
import sys
from io import BytesIO
from textwrap import dedent

from .. import errors, revision, shelf, shelf_ui, tests
from . import features, script


class ExpectShelver(shelf_ui.Shelver):
    """A variant of Shelver that intercepts console activity, for testing."""

    def __init__(self, work_tree, target_tree, diff_writer=None,
                 auto=False, auto_apply=False, file_list=None, message=None,
                 destroy=False, reporter=None):
        shelf_ui.Shelver.__init__(self, work_tree, target_tree, diff_writer,
                                  auto, auto_apply, file_list, message,
                                  destroy, reporter=reporter)
        self.expected = []
        self.diff_writer = BytesIO()

    def expect(self, message, response):
        self.expected.append((message, response))

    def prompt(self, message, choices, default):
        try:
            expected_message, response = self.expected.pop(0)
        except IndexError:
            raise AssertionError('Unexpected prompt: %s' % message)
        if message != expected_message:
            raise AssertionError('Wrong prompt: %s' % message)
        if choices != '&yes\n&No\n&finish\n&quit':
            raise AssertionError('Wrong choices: %s' % choices)
        return response


LINES_AJ = b'a\nb\nc\nd\ne\nf\ng\nh\ni\nj\n'


LINES_ZY = b'z\nb\nc\nd\ne\nf\ng\nh\ni\ny\n'


LINES_AY = b'a\nb\nc\nd\ne\nf\ng\nh\ni\ny\n'


class ShelfTestCase(tests.TestCaseWithTransport):

    def create_shelvable_tree(self):
        tree = self.make_branch_and_tree('tree')
        self.build_tree_contents([('tree/foo', LINES_AJ)])
        tree.add('foo', ids=b'foo-id')
        tree.commit('added foo')
        self.build_tree_contents([('tree/foo', LINES_ZY)])
        return tree


class TestShelver(ShelfTestCase):

    def test_unexpected_prompt_failure(self):
        tree = self.create_shelvable_tree()
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree())
        self.addCleanup(shelver.finalize)
        e = self.assertRaises(AssertionError, shelver.run)
        self.assertEqual('Unexpected prompt: Shelve?', str(e))

    def test_wrong_prompt_failure(self):
        tree = self.create_shelvable_tree()
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree())
        self.addCleanup(shelver.finalize)
        shelver.expect('foo', 0)
        e = self.assertRaises(AssertionError, shelver.run)
        self.assertEqual('Wrong prompt: Shelve?', str(e))

    def test_shelve_not_diff(self):
        tree = self.create_shelvable_tree()
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree())
        self.addCleanup(shelver.finalize)
        shelver.expect('Shelve?', 1)
        shelver.expect('Shelve?', 1)
        # No final shelving prompt because no changes were selected
        shelver.run()
        self.assertFileEqual(LINES_ZY, 'tree/foo')

    def test_shelve_diff_no(self):
        tree = self.create_shelvable_tree()
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree())
        self.addCleanup(shelver.finalize)
        shelver.expect('Shelve?', 0)
        shelver.expect('Shelve?', 0)
        shelver.expect('Shelve 2 change(s)?', 1)
        shelver.run()
        self.assertFileEqual(LINES_ZY, 'tree/foo')

    def test_shelve_diff(self):
        tree = self.create_shelvable_tree()
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree())
        self.addCleanup(shelver.finalize)
        shelver.expect('Shelve?', 0)
        shelver.expect('Shelve?', 0)
        shelver.expect('Shelve 2 change(s)?', 0)
        shelver.run()
        self.assertFileEqual(LINES_AJ, 'tree/foo')

    def test_shelve_one_diff(self):
        tree = self.create_shelvable_tree()
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree())
        self.addCleanup(shelver.finalize)
        shelver.expect('Shelve?', 0)
        shelver.expect('Shelve?', 1)
        shelver.expect('Shelve 1 change(s)?', 0)
        shelver.run()
        self.assertFileEqual(LINES_AY, 'tree/foo')

    def test_shelve_binary_change(self):
        tree = self.create_shelvable_tree()
        self.build_tree_contents([('tree/foo', b'\x00')])
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree())
        self.addCleanup(shelver.finalize)
        shelver.expect('Shelve binary changes?', 0)
        shelver.expect('Shelve 1 change(s)?', 0)
        shelver.run()
        self.assertFileEqual(LINES_AJ, 'tree/foo')

    def test_shelve_rename(self):
        tree = self.create_shelvable_tree()
        tree.rename_one('foo', 'bar')
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree())
        self.addCleanup(shelver.finalize)
        shelver.expect('Shelve renaming "foo" => "bar"?', 0)
        shelver.expect('Shelve?', 0)
        shelver.expect('Shelve?', 0)
        shelver.expect('Shelve 3 change(s)?', 0)
        shelver.run()
        self.assertFileEqual(LINES_AJ, 'tree/foo')

    def test_shelve_deletion(self):
        tree = self.create_shelvable_tree()
        os.unlink('tree/foo')
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree())
        self.addCleanup(shelver.finalize)
        shelver.expect('Shelve removing file "foo"?', 0)
        shelver.expect('Shelve 1 change(s)?', 0)
        shelver.run()
        self.assertFileEqual(LINES_AJ, 'tree/foo')

    def test_shelve_creation(self):
        tree = self.make_branch_and_tree('tree')
        tree.commit('add tree root')
        self.build_tree(['tree/foo'])
        tree.add('foo')
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree())
        self.addCleanup(shelver.finalize)
        shelver.expect('Shelve adding file "foo"?', 0)
        shelver.expect('Shelve 1 change(s)?', 0)
        shelver.run()
        self.assertPathDoesNotExist('tree/foo')

    def test_shelve_kind_change(self):
        tree = self.create_shelvable_tree()
        os.unlink('tree/foo')
        os.mkdir('tree/foo')
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree())
        self.addCleanup(shelver.finalize)
        shelver.expect('Shelve changing "foo" from file to directory?',
                       0)
        shelver.expect('Shelve 1 change(s)?', 0)

    def test_shelve_modify_target(self):
        self.requireFeature(features.SymlinkFeature(self.test_dir))
        tree = self.create_shelvable_tree()
        os.symlink('bar', 'tree/baz')
        tree.add('baz', ids=b'baz-id')
        tree.commit("Add symlink")
        os.unlink('tree/baz')
        os.symlink('vax', 'tree/baz')
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree())
        self.addCleanup(shelver.finalize)
        shelver.expect('Shelve changing target of "baz" from "bar" to '
                       '"vax"?', 0)
        shelver.expect('Shelve 1 change(s)?', 0)
        shelver.run()
        self.assertEqual('bar', os.readlink('tree/baz'))

    def test_shelve_finish(self):
        tree = self.create_shelvable_tree()
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree())
        self.addCleanup(shelver.finalize)
        shelver.expect('Shelve?', 2)
        shelver.expect('Shelve 2 change(s)?', 0)
        shelver.run()
        self.assertFileEqual(LINES_AJ, 'tree/foo')

    def test_shelve_quit(self):
        tree = self.create_shelvable_tree()
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree())
        self.addCleanup(shelver.finalize)
        shelver.expect('Shelve?', 3)
        self.assertRaises(errors.UserAbort, shelver.run)
        self.assertFileEqual(LINES_ZY, 'tree/foo')

    def test_shelve_all(self):
        self.create_shelvable_tree()
        shelver = ExpectShelver.from_args(sys.stdout, all=True,
                                          directory='tree')
        try:
            shelver.run()
        finally:
            shelver.finalize()
        self.assertFileEqual(LINES_AJ, 'tree/foo')

    def test_shelve_filename(self):
        tree = self.create_shelvable_tree()
        self.build_tree(['tree/bar'])
        tree.add('bar')
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree(), file_list=['bar'])
        self.addCleanup(shelver.finalize)
        shelver.expect('Shelve adding file "bar"?', 0)
        shelver.expect('Shelve 1 change(s)?', 0)
        shelver.run()

    def test_shelve_destroy(self):
        tree = self.create_shelvable_tree()
        shelver = shelf_ui.Shelver.from_args(sys.stdout, all=True,
                                             directory='tree', destroy=True)
        self.addCleanup(shelver.finalize)
        shelver.run()
        self.assertIs(None, tree.get_shelf_manager().last_shelf())
        self.assertFileEqual(LINES_AJ, 'tree/foo')

    @staticmethod
    def shelve_all(tree, target_revision_id):
        tree.lock_write()
        try:
            target = tree.branch.repository.revision_tree(target_revision_id)
            shelver = shelf_ui.Shelver(tree, target, auto=True,
                                       auto_apply=True)
            try:
                shelver.run()
            finally:
                shelver.finalize()
        finally:
            tree.unlock()

    def test_shelve_old_root_preserved(self):
        tree1 = self.make_branch_and_tree('tree1')
        tree1.commit('add root')
        tree1_root_id = tree1.path2id('')
        tree2 = self.make_branch_and_tree('tree2')
        rev2 = tree2.commit('add root')
        self.assertNotEqual(tree1_root_id, tree2.path2id(''))
        tree1.merge_from_branch(tree2.branch,
                                from_revision=revision.NULL_REVISION)
        tree1.commit('merging in tree2')
        self.assertEqual(tree1_root_id, tree1.path2id(''))
        # This is essentially assertNotRaises(InconsistentDelta)
        # With testtools 0.9.9, it can be rewritten as:
        # with ExpectedException(AssertionError,
        #                        'InconsistentDelta not raised'):
        #     with ExpectedException(errors.InconsistentDelta, ''):
        #         self.shelve_all(tree1, rev2)
        e = self.assertRaises(AssertionError, self.assertRaises,
                              errors.InconsistentDelta, self.shelve_all, tree1,
                              rev2)
        self.assertContainsRe('InconsistentDelta not raised', str(e))

    def test_shelve_split(self):
        outer_tree = self.make_branch_and_tree('outer')
        outer_tree.commit('Add root')
        inner_tree = self.make_branch_and_tree('outer/inner')
        rev2 = inner_tree.commit('Add root')
        outer_tree.subsume(inner_tree)
        # This is essentially assertNotRaises(ValueError).
        # The ValueError is 'None is not a valid file id'.
        self.expectFailure('Cannot shelve a join back to the inner tree.',
                           self.assertRaises, AssertionError,
                           self.assertRaises, ValueError, self.shelve_all,
                           outer_tree, rev2)


class TestApplyReporter(ShelfTestCase):

    def test_shelve_not_diff(self):
        tree = self.create_shelvable_tree()
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree(),
                                reporter=shelf_ui.ApplyReporter())
        self.addCleanup(shelver.finalize)
        shelver.expect('Apply change?', 1)
        shelver.expect('Apply change?', 1)
        # No final shelving prompt because no changes were selected
        shelver.run()
        self.assertFileEqual(LINES_ZY, 'tree/foo')

    def test_shelve_diff_no(self):
        tree = self.create_shelvable_tree()
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree(),
                                reporter=shelf_ui.ApplyReporter())
        self.addCleanup(shelver.finalize)
        shelver.expect('Apply change?', 0)
        shelver.expect('Apply change?', 0)
        shelver.expect('Apply 2 change(s)?', 1)
        shelver.run()
        self.assertFileEqual(LINES_ZY, 'tree/foo')

    def test_shelve_diff(self):
        tree = self.create_shelvable_tree()
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree(),
                                reporter=shelf_ui.ApplyReporter())
        self.addCleanup(shelver.finalize)
        shelver.expect('Apply change?', 0)
        shelver.expect('Apply change?', 0)
        shelver.expect('Apply 2 change(s)?', 0)
        shelver.run()
        self.assertFileEqual(LINES_AJ, 'tree/foo')

    def test_shelve_binary_change(self):
        tree = self.create_shelvable_tree()
        self.build_tree_contents([('tree/foo', b'\x00')])
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree(),
                                reporter=shelf_ui.ApplyReporter())
        self.addCleanup(shelver.finalize)
        shelver.expect('Apply binary changes?', 0)
        shelver.expect('Apply 1 change(s)?', 0)
        shelver.run()
        self.assertFileEqual(LINES_AJ, 'tree/foo')

    def test_shelve_rename(self):
        tree = self.create_shelvable_tree()
        tree.rename_one('foo', 'bar')
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree(),
                                reporter=shelf_ui.ApplyReporter())
        self.addCleanup(shelver.finalize)
        shelver.expect('Rename "bar" => "foo"?', 0)
        shelver.expect('Apply change?', 0)
        shelver.expect('Apply change?', 0)
        shelver.expect('Apply 3 change(s)?', 0)
        shelver.run()
        self.assertFileEqual(LINES_AJ, 'tree/foo')

    def test_shelve_deletion(self):
        tree = self.create_shelvable_tree()
        os.unlink('tree/foo')
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree(),
                                reporter=shelf_ui.ApplyReporter())
        self.addCleanup(shelver.finalize)
        shelver.expect('Add file "foo"?', 0)
        shelver.expect('Apply 1 change(s)?', 0)
        shelver.run()
        self.assertFileEqual(LINES_AJ, 'tree/foo')

    def test_shelve_creation(self):
        tree = self.make_branch_and_tree('tree')
        tree.commit('add tree root')
        self.build_tree(['tree/foo'])
        tree.add('foo')
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree(),
                                reporter=shelf_ui.ApplyReporter())
        self.addCleanup(shelver.finalize)
        shelver.expect('Delete file "foo"?', 0)
        shelver.expect('Apply 1 change(s)?', 0)
        shelver.run()
        self.assertPathDoesNotExist('tree/foo')

    def test_shelve_kind_change(self):
        tree = self.create_shelvable_tree()
        os.unlink('tree/foo')
        os.mkdir('tree/foo')
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree(),
                                reporter=shelf_ui.ApplyReporter())
        self.addCleanup(shelver.finalize)
        shelver.expect('Change "foo" from directory to a file?', 0)
        shelver.expect('Apply 1 change(s)?', 0)

    def test_shelve_modify_target(self):
        self.requireFeature(features.SymlinkFeature(self.test_dir))
        tree = self.create_shelvable_tree()
        os.symlink('bar', 'tree/baz')
        tree.add('baz', ids=b'baz-id')
        tree.commit("Add symlink")
        os.unlink('tree/baz')
        os.symlink('vax', 'tree/baz')
        tree.lock_tree_write()
        self.addCleanup(tree.unlock)
        shelver = ExpectShelver(tree, tree.basis_tree(),
                                reporter=shelf_ui.ApplyReporter())
        self.addCleanup(shelver.finalize)
        shelver.expect('Change target of "baz" from "vax" to "bar"?',
                       0)
        shelver.expect('Apply 1 change(s)?', 0)
        shelver.run()
        self.assertEqual('bar', os.readlink('tree/baz'))


class TestUnshelver(tests.TestCaseWithTransport):

    def create_tree_with_shelf(self):
        tree = self.make_branch_and_tree('tree')
        tree.lock_write()
        try:
            self.build_tree_contents([('tree/foo', LINES_AJ)])
            tree.add('foo', ids=b'foo-id')
            tree.commit('added foo')
            self.build_tree_contents([('tree/foo', LINES_ZY)])
            shelver = shelf_ui.Shelver(tree, tree.basis_tree(),
                                       auto_apply=True, auto=True)
            try:
                shelver.run()
            finally:
                shelver.finalize()
        finally:
            tree.unlock()
        return tree

    def test_unshelve(self):
        tree = self.create_tree_with_shelf()
        tree.lock_write()
        self.addCleanup(tree.unlock)
        manager = tree.get_shelf_manager()
        shelf_ui.Unshelver(tree, manager, 1, True, True, True).run()
        self.assertFileEqual(LINES_ZY, 'tree/foo')

    def test_unshelve_args(self):
        tree = self.create_tree_with_shelf()
        unshelver = shelf_ui.Unshelver.from_args(directory='tree')
        try:
            unshelver.run()
        finally:
            unshelver.tree.unlock()
        self.assertFileEqual(LINES_ZY, 'tree/foo')
        self.assertIs(None, tree.get_shelf_manager().last_shelf())

    def test_unshelve_args_dry_run(self):
        tree = self.create_tree_with_shelf()
        unshelver = shelf_ui.Unshelver.from_args(directory='tree',
                                                 action='dry-run')
        try:
            unshelver.run()
        finally:
            unshelver.tree.unlock()
        self.assertFileEqual(LINES_AJ, 'tree/foo')
        self.assertEqual(1, tree.get_shelf_manager().last_shelf())

    def test_unshelve_args_preview(self):
        tree = self.create_tree_with_shelf()
        write_diff_to = BytesIO()
        unshelver = shelf_ui.Unshelver.from_args(
            directory='tree', action='preview', write_diff_to=write_diff_to)
        try:
            unshelver.run()
        finally:
            unshelver.tree.unlock()
        # The changes were not unshelved.
        self.assertFileEqual(LINES_AJ, 'tree/foo')
        self.assertEqual(1, tree.get_shelf_manager().last_shelf())

        # But the diff was written to write_diff_to.
        diff = write_diff_to.getvalue()
        expected = dedent("""\
            @@ -1,4 +1,4 @@
            -a
            +z
             b
             c
             d
            @@ -7,4 +7,4 @@
             g
             h
             i
            -j
            +y

            """)
        self.assertEqualDiff(expected.encode('utf-8'), diff[-len(expected):])

    def test_unshelve_args_delete_only(self):
        tree = self.make_branch_and_tree('tree')
        manager = tree.get_shelf_manager()
        shelf_file = manager.new_shelf()[1]
        try:
            shelf_file.write(b'garbage')
        finally:
            shelf_file.close()
        unshelver = shelf_ui.Unshelver.from_args(directory='tree',
                                                 action='delete-only')
        try:
            unshelver.run()
        finally:
            unshelver.tree.unlock()
        self.assertIs(None, manager.last_shelf())

    def test_unshelve_args_invalid_shelf_id(self):
        tree = self.make_branch_and_tree('tree')
        manager = tree.get_shelf_manager()
        shelf_file = manager.new_shelf()[1]
        try:
            shelf_file.write(b'garbage')
        finally:
            shelf_file.close()
        self.assertRaises(shelf.InvalidShelfId,
                          shelf_ui.Unshelver.from_args, directory='tree',
                          action='delete-only', shelf_id='foo')


class TestUnshelveScripts(TestUnshelver,
                          script.TestCaseWithTransportAndScript):

    def test_unshelve_messages_keep(self):
        self.create_tree_with_shelf()
        self.run_script("""
$ cd tree
$ brz unshelve --keep
2>Using changes with id "1".
2> M  foo
2>All changes applied successfully.
""")

    def test_unshelve_messages_delete(self):
        self.create_tree_with_shelf()
        self.run_script("""
$ cd tree
$ brz unshelve --delete-only
2>Deleted changes with id "1".
""")

    def test_unshelve_messages_apply(self):
        self.create_tree_with_shelf()
        self.run_script("""
$ cd tree
$ brz unshelve --apply
2>Using changes with id "1".
2> M  foo
2>All changes applied successfully.
2>Deleted changes with id "1".
""")

    def test_unshelve_messages_dry_run(self):
        self.create_tree_with_shelf()
        self.run_script("""
$ cd tree
$ brz unshelve --dry-run
2>Using changes with id "1".
2> M  foo
""")



# Copyright (C) 2005-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for version_info"""

import os
import re
from io import BytesIO, StringIO

import yaml

from .. import registry, tests, version_info_formats
from ..bzr.rio import read_stanzas
from ..version_info_formats.format_custom import (CustomVersionInfoBuilder,
                                                  MissingTemplateVariable,
                                                  NoTemplate)
from ..version_info_formats.format_python import PythonVersionInfoBuilder
from ..version_info_formats.format_rio import RioVersionInfoBuilder
from ..version_info_formats.format_yaml import YamlVersionInfoBuilder
from . import TestCaseWithTransport


class VersionInfoTestCase(TestCaseWithTransport):

    def create_branch(self):
        wt = self.make_branch_and_tree('branch')

        self.build_tree(['branch/a'])
        wt.add('a')
        wt.commit('a', rev_id=b'r1')

        self.build_tree(['branch/b'])
        wt.add('b')
        wt.commit('b', rev_id=b'r2')

        self.build_tree_contents([('branch/a', b'new contents\n')])
        wt.commit('\xe52', rev_id=b'r3')

        return wt

    def create_tree_with_dotted_revno(self):
        wt = self.make_branch_and_tree('branch')
        self.build_tree(['branch/a'])
        wt.add('a')
        wt.commit('a', rev_id=b'r1')

        other = wt.controldir.sprout('other').open_workingtree()
        self.build_tree(['other/b.a'])
        other.add(['b.a'])
        other.commit('b.a', rev_id=b'o2')

        os.chdir('branch')
        self.run_bzr('merge ../other')
        wt.commit('merge', rev_id=b'merge')

        wt.update(revision=b'o2')

        return wt


class TestVersionInfoRio(VersionInfoTestCase):

    def test_rio_null(self):
        wt = self.make_branch_and_tree('branch')

        bio = BytesIO()
        builder = RioVersionInfoBuilder(wt.branch, working_tree=wt)
        builder.generate(bio)
        val = bio.getvalue()
        self.assertContainsRe(val, b'build-date:')
        self.assertContainsRe(val, b'revno: 0')

    def test_rio_dotted_revno(self):
        wt = self.create_tree_with_dotted_revno()

        bio = BytesIO()
        builder = RioVersionInfoBuilder(wt.branch, working_tree=wt)
        builder.generate(bio)
        val = bio.getvalue()
        self.assertContainsRe(val, b'revno: 1.1.1')

    def regen_text(self, wt, **kwargs):
        bio = BytesIO()
        builder = RioVersionInfoBuilder(wt.branch, working_tree=wt, **kwargs)
        builder.generate(bio)
        val = bio.getvalue()
        return val

    def test_simple(self):
        wt = self.create_branch()

        val = self.regen_text(wt)
        self.assertContainsRe(val, b'build-date:')
        self.assertContainsRe(val, b'date:')
        self.assertContainsRe(val, b'revno: 3')
        self.assertContainsRe(val, b'revision-id: r3')

    def test_clean(self):
        wt = self.create_branch()
        val = self.regen_text(wt, check_for_clean=True)
        self.assertContainsRe(val, b'clean: True')

    def test_no_clean(self):
        wt = self.create_branch()
        self.build_tree(['branch/c'])
        val = self.regen_text(wt, check_for_clean=True)
        self.assertContainsRe(val, b'clean: False')

    def test_history(self):
        wt = self.create_branch()

        val = self.regen_text(wt, include_revision_history=True)
        self.assertContainsRe(val, b'id: r1')
        self.assertContainsRe(val, b'message: a')
        self.assertContainsRe(val, b'id: r2')
        self.assertContainsRe(val, b'message: b')
        self.assertContainsRe(val, b'id: r3')
        self.assertContainsRe(val, b'message: \xc3\xa52')

    def regen(self, wt, **kwargs):
        bio = BytesIO()
        builder = RioVersionInfoBuilder(wt.branch, working_tree=wt, **kwargs)
        builder.generate(bio)
        bio.seek(0)
        stanzas = list(read_stanzas(bio))
        self.assertEqual(1, len(stanzas))
        return stanzas[0]

    def test_rio_version_hook(self):
        def update_stanza(rev, stanza):
            stanza.add('bla', 'bloe')
        RioVersionInfoBuilder.hooks.install_named_hook(
            'revision', update_stanza, None)
        wt = self.create_branch()

        stanza = self.regen(wt)
        self.assertEqual(['bloe'], stanza.get_all('bla'))

    def get_one_stanza(self, stanza, key):
        new_stanzas = list(read_stanzas(BytesIO(stanza[key].encode('utf8'))))
        self.assertEqual(1, len(new_stanzas))
        return new_stanzas[0]

    def test_build_date(self):
        wt = self.create_branch()
        stanza = self.regen(wt)
        self.assertTrue('date' in stanza)
        self.assertTrue('build-date' in stanza)
        self.assertEqual(['3'], stanza.get_all('revno'))
        self.assertEqual(['r3'], stanza.get_all('revision-id'))

    def test_not_clean(self):
        wt = self.create_branch()
        self.build_tree(['branch/c'])
        stanza = self.regen(wt, check_for_clean=True,
                            include_file_revisions=True)
        self.assertEqual(['False'], stanza.get_all('clean'))

    def test_file_revisions(self):
        wt = self.create_branch()
        self.build_tree(['branch/c'])
        stanza = self.regen(wt, check_for_clean=True,
                            include_file_revisions=True)
        # This assumes it's being run against a tree that does not update the
        # root revision on every commit.
        file_rev_stanza = self.get_one_stanza(stanza, 'file-revisions')
        self.assertEqual(['', 'a', 'b', 'c'], file_rev_stanza.get_all('path'))
        self.assertEqual(['r1', 'r3', 'r2', 'unversioned'],
                         file_rev_stanza.get_all('revision'))

    def test_revision_history(self):
        wt = self.create_branch()
        stanza = self.regen(wt, include_revision_history=True)
        revision_stanza = self.get_one_stanza(stanza, 'revisions')
        self.assertEqual(['r1', 'r2', 'r3'], revision_stanza.get_all('id'))
        self.assertEqual(['a', 'b', '\xe52'],
                         revision_stanza.get_all('message'))
        self.assertEqual(3, len(revision_stanza.get_all('date')))

    def test_file_revisions_with_rename(self):
        # a was modified, so it should show up modified again
        wt = self.create_branch()
        self.build_tree(['branch/a', 'branch/c'])
        wt.add('c')
        wt.rename_one('b', 'd')
        stanza = self.regen(wt, check_for_clean=True,
                            include_file_revisions=True)
        file_rev_stanza = self.get_one_stanza(stanza, 'file-revisions')
        self.assertEqual(['', 'a', 'b', 'c', 'd'],
                         file_rev_stanza.get_all('path'))
        self.assertEqual(['r1', 'modified', 'renamed to d', 'new',
                          'renamed from b'],
                         file_rev_stanza.get_all('revision'))

    def test_file_revisions_with_removal(self):
        wt = self.create_branch()
        self.build_tree(['branch/a', 'branch/c'])
        wt.add('c')
        wt.rename_one('b', 'd')

        wt.commit('modified', rev_id=b'r4')

        wt.remove(['c', 'd'])
        os.remove('branch/d')
        stanza = self.regen(wt, check_for_clean=True,
                            include_file_revisions=True)
        file_rev_stanza = self.get_one_stanza(stanza, 'file-revisions')
        self.assertEqual(['', 'a', 'c', 'd'], file_rev_stanza.get_all('path'))
        self.assertEqual(['r1', 'r4', 'unversioned', 'removed'],
                         file_rev_stanza.get_all('revision'))

    def test_revision(self):
        wt = self.create_branch()
        self.build_tree(['branch/a', 'branch/c'])
        wt.add('c')
        wt.rename_one('b', 'd')

        stanza = self.regen(
            wt, check_for_clean=True, include_file_revisions=True,
            revision_id=wt.last_revision())
        file_rev_stanza = self.get_one_stanza(stanza, 'file-revisions')
        self.assertEqual(['', 'a', 'b'], file_rev_stanza.get_all('path'))
        self.assertEqual(['r1', 'r3', 'r2'],
                         file_rev_stanza.get_all('revision'))


class TestVersionInfoYaml(VersionInfoTestCase):

    def test_yaml_null(self):
        wt = self.make_branch_and_tree('branch')

        bio = StringIO()
        builder = YamlVersionInfoBuilder(wt.branch, working_tree=wt)
        builder.generate(bio)
        val = bio.getvalue()
        self.assertContainsRe(val, 'build-date:')
        self.assertContainsRe(val, 'revno: \'0\'')

    def test_yaml_dotted_revno(self):
        wt = self.create_tree_with_dotted_revno()

        bio = StringIO()
        builder = YamlVersionInfoBuilder(wt.branch, working_tree=wt)
        builder.generate(bio)
        val = bio.getvalue()
        self.assertContainsRe(val, 'revno: 1.1.1')

    def regen_text(self, wt, **kwargs):
        bio = StringIO()
        builder = YamlVersionInfoBuilder(wt.branch, working_tree=wt, **kwargs)
        builder.generate(bio)
        val = bio.getvalue()
        return val

    def test_simple(self):
        wt = self.create_branch()

        val = self.regen_text(wt)
        self.assertContainsRe(val, 'build-date:')
        self.assertContainsRe(val, 'date:')
        self.assertContainsRe(val, 'revno: \'3\'')
        self.assertContainsRe(val, 'revision-id: r3')

    def test_clean(self):
        wt = self.create_branch()
        val = self.regen_text(wt, check_for_clean=True)
        self.assertContainsRe(val, 'clean: true')

    def test_no_clean(self):
        wt = self.create_branch()
        self.build_tree(['branch/c'])
        val = self.regen_text(wt, check_for_clean=True)
        self.assertContainsRe(val, 'clean: false')

    def test_history(self):
        wt = self.create_branch()

        val = self.regen_text(wt, include_revision_history=True)
        self.assertContainsRe(val, 'id: r1')
        self.assertContainsRe(val, 'message: a')
        self.assertContainsRe(val, 'id: r2')
        self.assertContainsRe(val, 'message: ')
        self.assertContainsRe(val, 'id: r3')
        self.assertContainsRe(val, re.escape('message: \"\\xE52\"'))

    def regen(self, wt, **kwargs):
        bio = StringIO()
        builder = YamlVersionInfoBuilder(wt.branch, working_tree=wt, **kwargs)
        builder.generate(bio)
        bio.seek(0)
        return yaml.safe_load(bio)

    def test_yaml_version_hook(self):
        def update_stanza(rev, stanza):
            stanza['bla'] = 'bloe'
        YamlVersionInfoBuilder.hooks.install_named_hook(
            'revision', update_stanza, None)
        wt = self.create_branch()

        stanza = self.regen(wt)
        self.assertEqual('bloe', stanza['bla'])

    def test_build_date(self):
        wt = self.create_branch()
        stanza = self.regen(wt)
        self.assertTrue('date' in stanza)
        self.assertTrue('build-date' in stanza)
        self.assertEqual('3', stanza['revno'])
        self.assertEqual('r3', stanza['revision-id'])

    def test_not_clean(self):
        wt = self.create_branch()
        self.build_tree(['branch/c'])
        stanza = self.regen(wt, check_for_clean=True,
                            include_file_revisions=True)
        self.assertEqual(False, stanza['clean'])

    def test_file_revisions(self):
        wt = self.create_branch()
        self.build_tree(['branch/c'])
        stanza = self.regen(wt, check_for_clean=True,
                            include_file_revisions=True)
        # This assumes it's being run against a tree that does not update the
        # root revision on every commit.
        file_rev_stanza = stanza['file-revisions']
        self.assertEqual(['', 'a', 'b', 'c'], [r['path'] for r in file_rev_stanza])
        self.assertEqual(['r1', 'r3', 'r2', 'unversioned'],
                         [r['revision'] for r in file_rev_stanza])

    def test_revision_history(self):
        wt = self.create_branch()
        stanza = self.regen(wt, include_revision_history=True)
        revision_stanza = stanza['revisions']
        self.assertEqual(
            ['r1', 'r2', 'r3'], [r['id'] for r in revision_stanza])
        self.assertEqual(
            ['a', 'b', '\xe52'],
            [r['message'] for r in revision_stanza])
        self.assertEqual(3, len([r['date'] for r in revision_stanza]))

    def test_file_revisions_with_rename(self):
        # a was modified, so it should show up modified again
        wt = self.create_branch()
        self.build_tree(['branch/a', 'branch/c'])
        wt.add('c')
        wt.rename_one('b', 'd')
        stanza = self.regen(wt, check_for_clean=True,
                            include_file_revisions=True)
        file_rev_stanza = stanza['file-revisions']
        self.assertEqual(['', 'a', 'b', 'c', 'd'],
                         [r['path'] for r in file_rev_stanza])
        self.assertEqual(['r1', 'modified', 'renamed to d', 'new',
                          'renamed from b'],
                         [r['revision'] for r in file_rev_stanza])

    def test_file_revisions_with_removal(self):
        wt = self.create_branch()
        self.build_tree(['branch/a', 'branch/c'])
        wt.add('c')
        wt.rename_one('b', 'd')

        wt.commit('modified', rev_id=b'r4')

        wt.remove(['c', 'd'])
        os.remove('branch/d')
        stanza = self.regen(wt, check_for_clean=True,
                            include_file_revisions=True)
        file_rev_stanza = stanza['file-revisions']
        self.assertEqual(['', 'a', 'c', 'd'], [r['path'] for r in file_rev_stanza])
        self.assertEqual(['r1', 'r4', 'unversioned', 'removed'],
                         [r['revision'] for r in file_rev_stanza])

    def test_revision(self):
        wt = self.create_branch()
        self.build_tree(['branch/a', 'branch/c'])
        wt.add('c')
        wt.rename_one('b', 'd')

        stanza = self.regen(
            wt, check_for_clean=True, include_file_revisions=True,
            revision_id=wt.last_revision())
        file_rev_stanza = stanza['file-revisions']
        self.assertEqual(['', 'a', 'b'], [r['path'] for r in file_rev_stanza])
        self.assertEqual(['r1', 'r3', 'r2'],
                         [r['revision'] for r in file_rev_stanza])

    def test_no_wt(self):
        wt = self.create_branch()
        self.build_tree(['branch/a', 'branch/c'])
        wt.add('c')
        wt.rename_one('b', 'd')

        bio = StringIO()
        builder = YamlVersionInfoBuilder(
            wt.branch, working_tree=None, check_for_clean=True,
            include_file_revisions=True, revision_id=None)
        builder.generate(bio)
        bio.seek(0)
        stanza = yaml.safe_load(bio)
        self.assertEqual([], stanza['file-revisions'])


class PythonVersionInfoTests(VersionInfoTestCase):

    def test_python_null(self):
        wt = self.make_branch_and_tree('branch')

        sio = StringIO()
        builder = PythonVersionInfoBuilder(wt.branch, working_tree=wt)
        builder.generate(sio)
        val = sio.getvalue()
        self.assertContainsRe(val, "'revision_id': None")
        self.assertContainsRe(val, "'revno': '0'")
        self.assertNotContainsString(val, '\n\n\n\n')

    def test_python_dotted_revno(self):
        wt = self.create_tree_with_dotted_revno()

        sio = StringIO()
        builder = PythonVersionInfoBuilder(wt.branch, working_tree=wt)
        builder.generate(sio)
        val = sio.getvalue()
        self.assertContainsRe(val, "'revno': '1.1.1'")

    def regen(self, wt, **kwargs):
        """Create a test module, import and return it"""
        builder = PythonVersionInfoBuilder(wt.branch, working_tree=wt,
                                           **kwargs)
        outf = StringIO()
        builder.generate(outf)
        local_vars = {}
        exec(outf.getvalue(), {}, local_vars)
        return local_vars

    def test_python_version(self):
        wt = self.create_branch()

        tvi = self.regen(wt)
        self.assertEqual('3', tvi['version_info']['revno'])
        self.assertEqual(b'r3', tvi['version_info']['revision_id'])
        self.assertTrue('date' in tvi['version_info'])
        self.assertEqual(None, tvi['version_info']['clean'])

        tvi = self.regen(wt, check_for_clean=True)
        self.assertTrue(tvi['version_info']['clean'])

        self.build_tree(['branch/c'])
        tvi = self.regen(wt, check_for_clean=True, include_file_revisions=True)
        self.assertFalse(tvi['version_info']['clean'])
        self.assertEqual(['', 'a', 'b', 'c'],
                         sorted(tvi['file_revisions'].keys()))
        self.assertEqual(b'r3', tvi['file_revisions']['a'])
        self.assertEqual(b'r2', tvi['file_revisions']['b'])
        self.assertEqual('unversioned', tvi['file_revisions']['c'])
        os.remove('branch/c')

        tvi = self.regen(wt, include_revision_history=True)

        rev_info = [(rev, message) for rev, message, timestamp, timezone
                    in tvi['revisions']]
        self.assertEqual([(b'r1', 'a'), (b'r2', 'b'),
                          (b'r3', '\xe52')], rev_info)

        # a was modified, so it should show up modified again
        self.build_tree(['branch/a', 'branch/c'])
        wt.add('c')
        wt.rename_one('b', 'd')
        tvi = self.regen(wt, check_for_clean=True, include_file_revisions=True)
        self.assertEqual(['', 'a', 'b', 'c', 'd'],
                         sorted(tvi['file_revisions'].keys()))
        self.assertEqual('modified', tvi['file_revisions']['a'])
        self.assertEqual('renamed to d', tvi['file_revisions']['b'])
        self.assertEqual('new', tvi['file_revisions']['c'])
        self.assertEqual('renamed from b', tvi['file_revisions']['d'])

        wt.commit('modified', rev_id=b'r4')
        wt.remove(['c', 'd'])
        os.remove('branch/d')
        tvi = self.regen(wt, check_for_clean=True, include_file_revisions=True)
        self.assertEqual(['', 'a', 'c', 'd'],
                         sorted(tvi['file_revisions'].keys()))
        self.assertEqual(b'r4', tvi['file_revisions']['a'])
        self.assertEqual('unversioned', tvi['file_revisions']['c'])
        self.assertEqual('removed', tvi['file_revisions']['d'])


class CustomVersionInfoTests(VersionInfoTestCase):

    def test_custom_null(self):
        sio = StringIO()
        wt = self.make_branch_and_tree('branch')
        builder = CustomVersionInfoBuilder(wt.branch, working_tree=wt,
                                           template='revno: {revno}')
        builder.generate(sio)
        self.assertEqual("revno: 0", sio.getvalue())

        builder = CustomVersionInfoBuilder(
            wt.branch, working_tree=wt,
            template='{revno} revid: {revision_id}')
        # revision_id is not available yet
        self.assertRaises(MissingTemplateVariable, builder.generate, sio)

    def test_custom_dotted_revno(self):
        sio = StringIO()
        wt = self.create_tree_with_dotted_revno()
        builder = CustomVersionInfoBuilder(
            wt.branch, working_tree=wt,
            template='{revno} revid: {revision_id}')
        builder.generate(sio)
        self.assertEqual("1.1.1 revid: o2", sio.getvalue())

    def regen(self, wt, tpl, **kwargs):
        sio = StringIO()
        builder = CustomVersionInfoBuilder(wt.branch, working_tree=wt,
                                           template=tpl, **kwargs)
        builder.generate(sio)
        val = sio.getvalue()
        return val

    def test_build_date(self):
        wt = self.create_branch()

        val = self.regen(wt, 'build-date: "{build_date}"\ndate: "{date}"')
        self.assertContainsRe(val, 'build-date: "[0-9-+: ]+"')
        self.assertContainsRe(val, 'date: "[0-9-+: ]+"')

    def test_revno(self):
        wt = self.create_branch()
        val = self.regen(wt, 'revno: {revno}')
        self.assertEqual(val, 'revno: 3')

    def test_revision_id(self):
        wt = self.create_branch()
        val = self.regen(wt, 'revision-id: {revision_id}')
        self.assertEqual(val, 'revision-id: r3')

    def test_clean(self):
        wt = self.create_branch()
        val = self.regen(wt, 'clean: {clean}', check_for_clean=True)
        self.assertEqual(val, 'clean: 1')

    def test_not_clean(self):
        wt = self.create_branch()

        self.build_tree(['branch/c'])
        val = self.regen(wt, 'clean: {clean}', check_for_clean=True)
        self.assertEqual(val, 'clean: 0')
        os.remove('branch/c')

    def test_custom_without_template(self):
        builder = CustomVersionInfoBuilder(None)
        sio = StringIO()
        self.assertRaises(NoTemplate, builder.generate, sio)


class TestBuilder(version_info_formats.VersionInfoBuilder):
    pass


class TestVersionInfoFormatRegistry(tests.TestCase):

    def setUp(self):
        super().setUp()
        self.overrideAttr(version_info_formats,
                          'format_registry', registry.Registry())

    def test_register_remove(self):
        registry = version_info_formats.format_registry
        registry.register('testbuilder',
                          TestBuilder, 'a simple test builder')
        self.assertIs(TestBuilder, registry.get('testbuilder'))
        self.assertEqual('a simple test builder',
                         registry.get_help('testbuilder'))
        registry.remove('testbuilder')
        self.assertRaises(KeyError, registry.get, 'testbuilder')



# Copyright (C) 2006, 2008-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for reconiliation behaviour that is repository independent."""


from breezy import errors, tests
from breezy.bzr import bzrdir
from breezy.reconcile import Reconciler, reconcile
from breezy.tests import per_repository


class TestWorksWithSharedRepositories(per_repository.TestCaseWithRepository):

    def test_reweave_empty(self):
        # we want a repo capable format
        parent = bzrdir.BzrDirMetaFormat1().initialize('.')
        parent.create_repository(shared=True)
        parent.root_transport.mkdir('child')
        child = bzrdir.BzrDirMetaFormat1().initialize('child')
        self.assertRaises(errors.NoRepositoryPresent, child.open_repository)
        reconciler = Reconciler(child)
        result = reconciler.reconcile()
        # smoke test for reconcile appears to work too.
        reconcile(child)
        # no inconsistent parents should have been found
        # but the values should have been set.
        self.assertEqual(0, result.inconsistent_parents)
        # and no garbage inventories
        self.assertEqual(0, result.garbage_inventories)


class TestReconciler(tests.TestCaseWithTransport):

    def test_reconciler_with_no_branch(self):
        repo = self.make_repository('repo')
        reconciler = Reconciler(repo.controldir)
        result = reconciler.reconcile()
        # no inconsistent parents should have been found
        # but the values should have been set.
        self.assertEqual(0, result.inconsistent_parents)
        # and no garbage inventories
        self.assertEqual(0, result.garbage_inventories)
        self.assertIs(None, result.fixed_branch_history)

    def test_reconciler_finds_branch(self):
        a_branch = self.make_branch('a_branch')
        reconciler = Reconciler(a_branch.controldir)
        result = reconciler.reconcile()

        # It should have checked the repository, and the branch
        self.assertEqual(0, result.inconsistent_parents)
        self.assertEqual(0, result.garbage_inventories)
        self.assertIs(False, result.fixed_branch_history)



# Copyright (C) 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import re

from .. import errors, gpg, mail_client, merge_directive, tests, trace

OUTPUT1 = b"""# Bazaar merge directive format 1
# revision_id: example:
# target_branch: http://example.com
# testament_sha1: sha
# timestamp: 1970-01-01 00:09:33 +0002
#\x20
booga"""

OUTPUT1_2 = b"""# Bazaar merge directive format 2 (Bazaar 0.90)
# revision_id: example:
# target_branch: http://example.com
# testament_sha1: sha
# timestamp: 1970-01-01 00:09:33 +0002
# base_revision_id: null:
#\x20
# Begin bundle
booga"""

OUTPUT2 = b"""# Bazaar merge directive format 1
# revision_id: example:
# target_branch: http://example.com
# testament_sha1: sha
# timestamp: 1970-01-01 00:09:33 +0002
# source_branch: http://example.org
# message: Hi mom!
#\x20
booga"""

OUTPUT2_2 = b"""# Bazaar merge directive format 2 (Bazaar 0.90)
# revision_id: example:
# target_branch: http://example.com
# testament_sha1: sha
# timestamp: 1970-01-01 00:09:33 +0002
# source_branch: http://example.org
# message: Hi mom!
# base_revision_id: null:
#\x20
# Begin patch
booga"""

INPUT1 = b"""
I was thinking today about creating a merge directive.

So I did.

Here it is.

(I've pasted it in the body of this message)

Aaron

# Bazaar merge directive format 1\r
# revision_id: example:
# target_branch: http://example.com
# testament_sha1: sha
# timestamp: 1970-01-01 00:09:33 +0002
# source_branch: http://example.org
# message: Hi mom!
#\x20
booga""".splitlines(True)


INPUT1_2 = b"""
I was thinking today about creating a merge directive.

So I did.

Here it is.

(I've pasted it in the body of this message)

Aaron

# Bazaar merge directive format 2 (Bazaar 0.90)\r
# revision_id: example:
# target_branch: http://example.com
# testament_sha1: sha
# timestamp: 1970-01-01 00:09:33 +0002
# source_branch: http://example.org
# base_revision_id: null:
# message: Hi mom!
#\x20
# Begin patch
booga""".splitlines(True)


INPUT1_2_OLD = b"""
I was thinking today about creating a merge directive.

So I did.

Here it is.

(I've pasted it in the body of this message)

Aaron

# Bazaar merge directive format 2 (Bazaar 0.19)\r
# revision_id: example:
# target_branch: http://example.com
# testament_sha1: sha
# timestamp: 1970-01-01 00:09:33 +0002
# source_branch: http://example.org
# base_revision_id: null:
# message: Hi mom!
#\x20
# Begin patch
booga""".splitlines(True)


OLD_DIRECTIVE_2 = b"""# Bazaar merge directive format 2 (Bazaar 0.19)
# revision_id: abentley@panoramicfeedback.com-20070807234458-\
#   nzhkoyza56lan7z5
# target_branch: http://panoramicfeedback.com/opensource/bzr/repo\
#   /bzr.ab
# testament_sha1: d825a5cdb267a90ec2ba86b00895f3d8a9bed6bf
# timestamp: 2007-08-10 16:15:02 -0400
# source_branch: http://panoramicfeedback.com/opensource/bzr/repo\
#   /bzr.ab
# base_revision_id: abentley@panoramicfeedback.com-20070731163346-\
#   623xwcycwij91xen
#
""".splitlines(True)


class TestMergeDirective:

    def test_merge_source(self):
        time = 500000.0
        timezone = 5 * 3600
        self.assertRaises(errors.NoMergeSource, self.make_merge_directive,
                          b'example:', b'sha', time, timezone, 'http://example.com')
        self.assertRaises(errors.NoMergeSource, self.make_merge_directive,
                          b'example:', b'sha', time, timezone, 'http://example.com',
                          patch_type='diff')
        self.make_merge_directive(b'example:', b'sha', time, timezone,
                                  'http://example.com', source_branch='http://example.org')
        md = self.make_merge_directive(b'null:', b'sha', time, timezone,
                                       'http://example.com', patch=b'blah', patch_type='bundle')
        self.assertIs(None, md.source_branch)
        md2 = self.make_merge_directive(b'null:', b'sha', time, timezone,
                                        'http://example.com', patch=b'blah', patch_type='bundle',
                                        source_branch='bar')
        self.assertEqual('bar', md2.source_branch)

    def test_serialization(self):
        time = 453
        timezone = 120
        md = self.make_merge_directive(b'example:', b'sha', time, timezone,
                                       'http://example.com', patch=b'booga', patch_type='bundle')
        self.assertEqualDiff(self.OUTPUT1, b''.join(md.to_lines()))
        md = self.make_merge_directive(b'example:', b'sha', time, timezone,
                                       'http://example.com', source_branch="http://example.org",
                                       patch=b'booga', patch_type='diff', message="Hi mom!")
        self.assertEqualDiff(self.OUTPUT2, b''.join(md.to_lines()))

    def test_deserialize_junk(self):
        time = 501
        self.assertRaises(errors.NotAMergeDirective,
                          merge_directive.MergeDirective.from_lines, [b'lala'])

    def test_deserialize_empty(self):
        self.assertRaises(errors.NotAMergeDirective,
                          merge_directive.MergeDirective.from_lines, [])

    def test_deserialize_leading_junk(self):
        md = merge_directive.MergeDirective.from_lines(self.INPUT1)
        self.assertEqual(b'example:', md.revision_id)
        self.assertEqual(b'sha', md.testament_sha1)
        self.assertEqual('http://example.com', md.target_branch)
        self.assertEqual('http://example.org', md.source_branch)
        self.assertEqual(453, md.time)
        self.assertEqual(120, md.timezone)
        self.assertEqual(b'booga', md.patch)
        self.assertEqual('diff', md.patch_type)
        self.assertEqual('Hi mom!', md.message)

    def test_roundtrip(self):
        time = 500000
        timezone = 7.5 * 3600
        md = self.make_merge_directive(b'example:', b'sha', time, timezone,
                                       'http://example.com', source_branch="http://example.org",
                                       patch=b'booga', patch_type='diff')
        md2 = merge_directive.MergeDirective.from_lines(md.to_lines())
        self.assertEqual(b'example:', md2.revision_id)
        self.assertIsInstance(md2.revision_id, bytes)
        self.assertEqual(b'sha', md2.testament_sha1)
        self.assertEqual('http://example.com', md2.target_branch)
        self.assertEqual('http://example.org', md2.source_branch)
        self.assertEqual(time, md2.time)
        self.assertEqual(timezone, md2.timezone)
        self.assertEqual('diff', md2.patch_type)
        self.assertEqual(b'booga', md2.patch)
        self.assertEqual(None, md2.message)
        self.set_bundle(md, b"# Bazaar revision bundle v0.9\n#\n")
        md.message = "Hi mom!"
        lines = md.to_lines()
        md3 = merge_directive.MergeDirective.from_lines(lines)
        self.assertEqual(b"# Bazaar revision bundle v0.9\n#\n", md3.bundle)
        self.assertEqual("bundle", md3.patch_type)
        self.assertContainsRe(md3.to_lines()[0],
                              b'^# Bazaar merge directive format ')
        self.assertEqual("Hi mom!", md3.message)
        md3.clear_payload()
        self.assertIs(None, md3.get_raw_bundle())
        md4 = merge_directive.MergeDirective.from_lines(md3.to_lines())
        self.assertIs(None, md4.patch_type)


class TestMergeDirective1(tests.TestCase, TestMergeDirective):
    """Test merge directive format 1"""

    INPUT1 = INPUT1

    OUTPUT1 = OUTPUT1

    OUTPUT2 = OUTPUT2

    def make_merge_directive(self, revision_id, testament_sha1, time, timezone,
                             target_branch, patch=None, patch_type=None,
                             source_branch=None, message=None):
        return merge_directive.MergeDirective(revision_id, testament_sha1,
                                              time, timezone, target_branch, patch, patch_type,
                                              source_branch, message)

    @staticmethod
    def set_bundle(md, value):
        md.patch = value

    def test_require_patch(self):
        time = 500.0
        timezone = 120
        self.assertRaises(errors.PatchMissing, merge_directive.MergeDirective,
                          b'example:', b'sha', time, timezone, 'http://example.com',
                          patch_type='bundle')
        md = merge_directive.MergeDirective(b'example:', b'sha1', time, timezone,
                                            'http://example.com', source_branch="http://example.org",
                                            patch=b'', patch_type='diff')
        self.assertEqual(md.patch, b'')


class TestMergeDirective2(tests.TestCase, TestMergeDirective):
    """Test merge directive format 2"""

    INPUT1 = INPUT1_2

    OUTPUT1 = OUTPUT1_2

    OUTPUT2 = OUTPUT2_2

    def make_merge_directive(self, revision_id, testament_sha1, time, timezone,
                             target_branch, patch=None, patch_type=None,
                             source_branch=None, message=None, base_revision_id=b'null:'):
        if patch_type == 'bundle':
            bundle = patch
            patch = None
        else:
            bundle = None
        return merge_directive.MergeDirective2(revision_id, testament_sha1,
                                               time, timezone, target_branch, patch, source_branch, message,
                                               bundle, base_revision_id)

    @staticmethod
    def set_bundle(md, value):
        md.bundle = value


EMAIL1 = """From: "J. Random Hacker" <jrandom@example.com>
Subject: Commit of rev2a
To: pqm@example.com
User-Agent: Bazaar \\(.*\\)

# Bazaar merge directive format 1
# revision_id: rev2a
# target_branch: (.|\n)*
# testament_sha1: .*
# timestamp: 1970-01-01 00:08:56 \\+0001
# source_branch: (.|\n)*
"""


EMAIL1_2 = """From: "J. Random Hacker" <jrandom@example.com>
Subject: Commit of rev2a
To: pqm@example.com
User-Agent: Bazaar \\(.*\\)

# Bazaar merge directive format 2 \\(Bazaar 0.90\\)
# revision_id: rev2a
# target_branch: (.|\n)*
# testament_sha1: .*
# timestamp: 1970-01-01 00:08:56 \\+0001
# source_branch: (.|\n)*
"""


EMAIL2 = """From: "J. Random Hacker" <jrandom@example.com>
Subject: Commit of rev2a with special message
To: pqm@example.com
User-Agent: Bazaar \\(.*\\)

# Bazaar merge directive format 1
# revision_id: rev2a
# target_branch: (.|\n)*
# testament_sha1: .*
# timestamp: 1970-01-01 00:08:56 \\+0001
# source_branch: (.|\n)*
# message: Commit of rev2a with special message
"""

EMAIL2_2 = """From: "J. Random Hacker" <jrandom@example.com>
Subject: Commit of rev2a with special message
To: pqm@example.com
User-Agent: Bazaar \\(.*\\)

# Bazaar merge directive format 2 \\(Bazaar 0.90\\)
# revision_id: rev2a
# target_branch: (.|\n)*
# testament_sha1: .*
# timestamp: 1970-01-01 00:08:56 \\+0001
# source_branch: (.|\n)*
# message: Commit of rev2a with special message
"""


class TestMergeDirectiveBranch:

    def make_trees(self):
        tree_a = self.make_branch_and_tree('tree_a')
        tree_a.branch.get_config_stack().set(
            'email', 'J. Random Hacker <jrandom@example.com>')
        self.build_tree_contents([('tree_a/file', b'content_a\ncontent_b\n'),
                                  ('tree_a/file_2', b'content_x\rcontent_y\r')])
        tree_a.add(['file', 'file_2'])
        tree_a.commit('message', rev_id=b'rev1')
        tree_b = tree_a.controldir.sprout('tree_b').open_workingtree()
        branch_c = tree_a.controldir.sprout('branch_c').open_branch()
        tree_b.commit('message', rev_id=b'rev2b')
        self.build_tree_contents([('tree_a/file', b'content_a\ncontent_c \n'),
                                  ('tree_a/file_2', b'content_x\rcontent_z\r')])
        tree_a.commit('Commit of rev2a', rev_id=b'rev2a')
        return tree_a, tree_b, branch_c

    def test_empty_target(self):
        tree_a, tree_b, branch_c = self.make_trees()
        tree_d = self.make_branch_and_tree('tree_d')
        md2 = self.from_objects(tree_a.branch.repository, b'rev2a', 500, 120,
                                tree_d.branch.base, patch_type='diff',
                                public_branch=tree_a.branch.base)

    def test_disk_name(self):
        tree_a, tree_b, branch_c = self.make_trees()
        tree_a.branch.nick = 'fancy <name>'
        md = self.from_objects(tree_a.branch.repository, b'rev2a', 500, 120,
                               tree_b.branch.base)
        self.assertEqual('fancy-name-2', md.get_disk_name(tree_a.branch))

    def test_disk_name_old_revno(self):
        tree_a, tree_b, branch_c = self.make_trees()
        tree_a.branch.nick = 'fancy-name'
        md = self.from_objects(tree_a.branch.repository, b'rev1', 500, 120,
                               tree_b.branch.base)
        self.assertEqual('fancy-name-1', md.get_disk_name(tree_a.branch))

    def test_generate_patch(self):
        tree_a, tree_b, branch_c = self.make_trees()
        md2 = self.from_objects(tree_a.branch.repository, b'rev2a', 500, 120,
                                tree_b.branch.base, patch_type='diff',
                                public_branch=tree_a.branch.base)
        self.assertNotContainsRe(md2.patch, b'Bazaar revision bundle')
        self.assertContainsRe(md2.patch, b'\\+content_c')
        self.assertNotContainsRe(md2.patch, b'\\+\\+\\+ b/')
        self.assertContainsRe(md2.patch, b'\\+\\+\\+ file')

    def test_public_branch(self):
        tree_a, tree_b, branch_c = self.make_trees()
        self.assertRaises(errors.PublicBranchOutOfDate,
                          self.from_objects, tree_a.branch.repository, b'rev2a', 500, 144,
                          tree_b.branch.base, public_branch=branch_c.base, patch_type='diff')
        self.assertRaises(errors.PublicBranchOutOfDate,
                          self.from_objects, tree_a.branch.repository, b'rev2a', 500, 144,
                          tree_b.branch.base, public_branch=branch_c.base, patch_type=None)
        # public branch is not checked if patch format is bundle.
        md1 = self.from_objects(tree_a.branch.repository, b'rev2a', 500, 144,
                                tree_b.branch.base, public_branch=branch_c.base)
        # public branch is provided with a bundle, despite possibly being out
        # of date, because it's not required if a bundle is present.
        self.assertEqual(md1.source_branch, branch_c.base)
        # Once we update the public branch, we can generate a diff.
        branch_c.pull(tree_a.branch)
        md3 = self.from_objects(tree_a.branch.repository, b'rev2a', 500, 144,
                                tree_b.branch.base, patch_type=None, public_branch=branch_c.base)

    def test_use_public_submit_branch(self):
        tree_a, tree_b, branch_c = self.make_trees()
        branch_c.pull(tree_a.branch)
        md = self.from_objects(tree_a.branch.repository, b'rev2a', 500, 144,
                               tree_b.branch.base, patch_type=None, public_branch=branch_c.base)
        self.assertEqual(md.target_branch, tree_b.branch.base)
        tree_b.branch.set_public_branch('http://example.com')
        md2 = self.from_objects(
            tree_a.branch.repository, b'rev2a', 500, 144, tree_b.branch.base,
            patch_type=None, public_branch=branch_c.base)
        self.assertEqual(md2.target_branch, 'http://example.com')

    def test_message(self):
        tree_a, tree_b, branch_c = self.make_trees()
        md3 = self.from_objects(tree_a.branch.repository, b'rev1', 500, 120,
                                tree_b.branch.base, patch_type=None, public_branch=branch_c.base,
                                message='Merge message')
        md3.to_lines()
        self.assertIs(None, md3.patch)
        self.assertEqual('Merge message', md3.message)

    def test_generate_bundle(self):
        tree_a, tree_b, branch_c = self.make_trees()
        md1 = self.from_objects(tree_a.branch.repository, b'rev2a', 500, 120,
                                tree_b.branch.base, public_branch=branch_c.base)

        self.assertContainsRe(md1.get_raw_bundle(), b'Bazaar revision bundle')
        self.assertContainsRe(md1.patch, b'\\+content_c')
        self.assertNotContainsRe(md1.patch, b'\\+content_a')
        self.assertContainsRe(md1.patch, b'\\+content_c')
        self.assertNotContainsRe(md1.patch, b'\\+content_a')

    def test_broken_bundle(self):
        tree_a, tree_b, branch_c = self.make_trees()
        md1 = self.from_objects(tree_a.branch.repository, b'rev2a', 500, 120,
                                tree_b.branch.base, public_branch=branch_c.base)
        lines = md1.to_lines()
        lines = [l.replace(b'\n', b'\r\n') for l in lines]
        md2 = merge_directive.MergeDirective.from_lines(lines)
        self.assertEqual(b'rev2a', md2.revision_id)

    def test_signing(self):
        time = 453
        timezone = 7200

        class FakeBranch:
            def get_config_stack(self):
                return self
        md = self.make_merge_directive(b'example:', b'sha', time, timezone,
                                       'http://example.com', source_branch="http://example.org",
                                       patch=b'booga', patch_type='diff')
        old_strategy = gpg.GPGStrategy
        gpg.GPGStrategy = gpg.LoopbackGPGStrategy
        try:
            signed = md.to_signed(FakeBranch())
        finally:
            gpg.GPGStrategy = old_strategy
        self.assertContainsRe(signed, b'^-----BEGIN PSEUDO-SIGNED CONTENT')
        self.assertContainsRe(signed, b'example.org')
        self.assertContainsRe(signed, b'booga')

    def test_email(self):
        tree_a, tree_b, branch_c = self.make_trees()
        md = self.from_objects(tree_a.branch.repository, b'rev2a', 476, 60,
                               tree_b.branch.base, patch_type=None,
                               public_branch=tree_a.branch.base)
        message = md.to_email('pqm@example.com', tree_a.branch)
        self.assertContainsRe(message.as_string(), self.EMAIL1)
        md.message = 'Commit of rev2a with special message'
        message = md.to_email('pqm@example.com', tree_a.branch)
        self.assertContainsRe(message.as_string(), self.EMAIL2)

    def test_install_revisions_branch(self):
        tree_a, tree_b, branch_c = self.make_trees()
        md = self.from_objects(tree_a.branch.repository, b'rev2a', 500, 36,
                               tree_b.branch.base, patch_type=None,
                               public_branch=tree_a.branch.base)
        self.assertFalse(tree_b.branch.repository.has_revision(b'rev2a'))
        revision = md.install_revisions(tree_b.branch.repository)
        self.assertEqual(b'rev2a', revision)
        self.assertTrue(tree_b.branch.repository.has_revision(b'rev2a'))

    def test_get_merge_request(self):
        tree_a, tree_b, branch_c = self.make_trees()
        md = self.from_objects(tree_a.branch.repository, b'rev2a', 500, 36,
                               tree_b.branch.base, patch_type='bundle',
                               public_branch=tree_a.branch.base)
        self.assertFalse(tree_b.branch.repository.has_revision(b'rev2a'))
        md.install_revisions(tree_b.branch.repository)
        base, revision, verified = md.get_merge_request(
            tree_b.branch.repository)
        if isinstance(md, merge_directive.MergeDirective):
            self.assertIs(None, base)
            self.assertEqual('inapplicable', verified)
        else:
            self.assertEqual(b'rev1', base)
            self.assertEqual('verified', verified)
        self.assertEqual(b'rev2a', revision)
        self.assertTrue(tree_b.branch.repository.has_revision(b'rev2a'))
        md = self.from_objects(tree_a.branch.repository, b'rev2a', 500, 36,
                               tree_b.branch.base, patch_type=None,
                               public_branch=tree_a.branch.base)
        base, revision, verified = md.get_merge_request(
            tree_b.branch.repository)
        if isinstance(md, merge_directive.MergeDirective):
            self.assertIs(None, base)
            self.assertEqual('inapplicable', verified)
        else:
            self.assertEqual(b'rev1', base)
            self.assertEqual('inapplicable', verified)
        md = self.from_objects(tree_a.branch.repository, b'rev2a', 500, 36,
                               tree_b.branch.base, patch_type='diff',
                               public_branch=tree_a.branch.base)
        base, revision, verified = md.get_merge_request(
            tree_b.branch.repository)
        if isinstance(md, merge_directive.MergeDirective):
            self.assertIs(None, base)
            self.assertEqual('inapplicable', verified)
        else:
            self.assertEqual(b'rev1', base)
            self.assertEqual('verified', verified)
        md.patch = b'asdf'
        base, revision, verified = md.get_merge_request(
            tree_b.branch.repository)
        if isinstance(md, merge_directive.MergeDirective):
            self.assertIs(None, base)
            self.assertEqual('inapplicable', verified)
        else:
            self.assertEqual(b'rev1', base)
            self.assertEqual('failed', verified)

    def test_install_revisions_bundle(self):
        tree_a, tree_b, branch_c = self.make_trees()
        md = self.from_objects(tree_a.branch.repository, b'rev2a', 500, 36,
                               tree_b.branch.base, patch_type='bundle',
                               public_branch=tree_a.branch.base)
        self.assertFalse(tree_b.branch.repository.has_revision(b'rev2a'))
        revision = md.install_revisions(tree_b.branch.repository)
        self.assertEqual(b'rev2a', revision)
        self.assertTrue(tree_b.branch.repository.has_revision(b'rev2a'))

    def test_get_target_revision_nofetch(self):
        tree_a, tree_b, branch_c = self.make_trees()
        tree_b.branch.fetch(tree_a.branch)
        md = self.from_objects(tree_a.branch.repository, b'rev2a', 500, 36,
                               tree_b.branch.base, patch_type=None,
                               public_branch=tree_a.branch.base)
        md.source_branch = '/dev/null'
        revision = md.install_revisions(tree_b.branch.repository)
        self.assertEqual(b'rev2a', revision)

    def test_use_submit_for_missing_dependency(self):
        tree_a, tree_b, branch_c = self.make_trees()
        branch_c.pull(tree_a.branch)
        self.build_tree_contents([('tree_a/file', b'content_q\ncontent_r\n')])
        tree_a.commit('rev3a', rev_id=b'rev3a')
        md = self.from_objects(tree_a.branch.repository, b'rev3a', 500, 36,
                               branch_c.base, base_revision_id=b'rev2a')
        revision = md.install_revisions(tree_b.branch.repository)

    def test_handle_target_not_a_branch(self):
        tree_a, tree_b, branch_c = self.make_trees()
        branch_c.pull(tree_a.branch)
        self.build_tree_contents([('tree_a/file', b'content_q\ncontent_r\n')])
        tree_a.commit('rev3a', rev_id=b'rev3a')
        md = self.from_objects(tree_a.branch.repository, b'rev3a', 500, 36,
                               branch_c.base, base_revision_id=b'rev2a')
        md.target_branch = self.get_url('not-a-branch')
        self.assertRaises(errors.TargetNotBranch, md.install_revisions,
                          tree_b.branch.repository)


class TestMergeDirective1Branch(tests.TestCaseWithTransport,
                                TestMergeDirectiveBranch):
    """Test merge directive format 1 with a branch"""

    EMAIL1 = EMAIL1

    EMAIL2 = EMAIL2

    def from_objects(self, repository, revision_id, time, timezone,
                     target_branch, patch_type='bundle', local_target_branch=None,
                     public_branch=None, message=None, base_revision_id=None):
        if base_revision_id is not None:
            raise tests.TestNotApplicable('This format does not support'
                                          ' explicit bases.')
        with repository.lock_write():
            return merge_directive.MergeDirective.from_objects(repository,
                                                               revision_id, time, timezone, target_branch, patch_type,
                                                               local_target_branch, public_branch, message)

    def make_merge_directive(self, revision_id, testament_sha1, time, timezone,
                             target_branch, patch=None, patch_type=None,
                             source_branch=None, message=None):
        return merge_directive.MergeDirective(revision_id, testament_sha1,
                                              time, timezone, target_branch, patch, patch_type,
                                              source_branch, message)


class TestMergeDirective2Branch(tests.TestCaseWithTransport,
                                TestMergeDirectiveBranch):
    """Test merge directive format 2 with a branch"""

    EMAIL1 = EMAIL1_2

    EMAIL2 = EMAIL2_2

    def from_objects(self, repository, revision_id, time, timezone,
                     target_branch, patch_type='bundle', local_target_branch=None,
                     public_branch=None, message=None, base_revision_id=None):
        include_patch = (patch_type in ('bundle', 'diff'))
        include_bundle = (patch_type == 'bundle')
        self.assertTrue(patch_type in ('bundle', 'diff', None))
        return merge_directive.MergeDirective2.from_objects(
            repository, revision_id, time, timezone, target_branch,
            include_patch, include_bundle, local_target_branch, public_branch,
            message, base_revision_id)

    def make_merge_directive(self, revision_id, testament_sha1, time, timezone,
                             target_branch, patch=None, patch_type=None,
                             source_branch=None, message=None, base_revision_id=b'null:'):
        if patch_type == 'bundle':
            bundle = patch
            patch = None
        else:
            bundle = None
        return merge_directive.MergeDirective2(revision_id, testament_sha1,
                                               time, timezone, target_branch, patch, source_branch, message,
                                               bundle, base_revision_id)

    def test_base_revision(self):
        tree_a, tree_b, branch_c = self.make_trees()
        md = self.from_objects(tree_a.branch.repository, b'rev2a', 500, 60,
                               tree_b.branch.base, patch_type='bundle',
                               public_branch=tree_a.branch.base, base_revision_id=None)
        self.assertEqual(b'rev1', md.base_revision_id)
        md = self.from_objects(tree_a.branch.repository, b'rev2a', 500, 60,
                               tree_b.branch.base, patch_type='bundle',
                               public_branch=tree_a.branch.base, base_revision_id=b'null:')
        self.assertEqual(b'null:', md.base_revision_id)
        lines = md.to_lines()
        md2 = merge_directive.MergeDirective.from_lines(lines)
        self.assertEqual(md2.base_revision_id, md.base_revision_id)

    def test_patch_verification(self):
        tree_a, tree_b, branch_c = self.make_trees()
        md = self.from_objects(tree_a.branch.repository, b'rev2a', 500, 60,
                               tree_b.branch.base, patch_type='bundle',
                               public_branch=tree_a.branch.base)
        lines = md.to_lines()
        md2 = merge_directive.MergeDirective.from_lines(lines)
        md2._verify_patch(tree_a.branch.repository)
        # Strip trailing whitespace
        md2.patch = md2.patch.replace(b' \n', b'\n')
        md2._verify_patch(tree_a.branch.repository)
        # Convert to Mac line-endings
        md2.patch = re.sub(b'(\r\n|\r|\n)', b'\r', md2.patch)
        self.assertTrue(md2._verify_patch(tree_a.branch.repository))
        # Convert to DOS line-endings
        md2.patch = re.sub(b'(\r\n|\r|\n)', b'\r\n', md2.patch)
        self.assertTrue(md2._verify_patch(tree_a.branch.repository))
        md2.patch = md2.patch.replace(b'content_c', b'content_d')
        self.assertFalse(md2._verify_patch(tree_a.branch.repository))


class TestParseOldMergeDirective2(tests.TestCase):

    def test_parse_old_merge_directive(self):
        md = merge_directive.MergeDirective.from_lines(INPUT1_2_OLD)
        self.assertEqual(b'example:', md.revision_id)
        self.assertEqual(b'sha', md.testament_sha1)
        self.assertEqual('http://example.com', md.target_branch)
        self.assertEqual('http://example.org', md.source_branch)
        self.assertEqual(453, md.time)
        self.assertEqual(120, md.timezone)
        self.assertEqual(b'booga', md.patch)
        self.assertEqual('diff', md.patch_type)
        self.assertEqual('Hi mom!', md.message)


class TestHook:
    """Hook callback for test purposes."""

    def __init__(self, result=None):
        self.calls = []
        self.result = result

    def __call__(self, params):
        self.calls.append(params)
        return self.result


class HookMailClient(mail_client.MailClient):
    """Mail client for testing hooks."""

    def __init__(self, config):
        self.body = None
        self.config = config

    def compose(self, prompt, to, subject, attachment, mime_subtype,
                extension, basename=None, body=None):
        self.body = body


class TestBodyHook(tests.TestCaseWithTransport):

    def compose_with_hooks(self, test_hooks, supports_body=True):
        client = HookMailClient({})
        client.supports_body = supports_body
        for test_hook in test_hooks:
            merge_directive.MergeDirective.hooks.install_named_hook(
                'merge_request_body', test_hook, 'test')
        tree = self.make_branch_and_tree('foo')
        tree.commit('foo')
        directive = merge_directive.MergeDirective2(
            tree.branch.last_revision(), b'sha', 0, 0, b'sha',
            source_branch=tree.branch.base,
            base_revision_id=tree.branch.last_revision(),
            message='This code rox')
        directive.compose_merge_request(client, 'jrandom@example.com',
                                        None, tree.branch)
        return client, directive

    def test_no_supports_body(self):
        test_hook = TestHook('foo')
        old_warn = trace.warning
        warnings = []

        def warn(*args):
            warnings.append(args)
        trace.warning = warn
        try:
            client, directive = self.compose_with_hooks([test_hook],
                                                        supports_body=False)
        finally:
            trace.warning = old_warn
        self.assertEqual(0, len(test_hook.calls))
        self.assertEqual(('Cannot run merge_request_body hooks because mail'
                          ' client %s does not support message bodies.',
                          'HookMailClient'), warnings[0])

    def test_body_hook(self):
        test_hook = TestHook('foo')
        client, directive = self.compose_with_hooks([test_hook])
        self.assertEqual(1, len(test_hook.calls))
        self.assertEqual('foo', client.body)
        params = test_hook.calls[0]
        self.assertIsInstance(params,
                              merge_directive.MergeRequestBodyParams)
        self.assertIs(None, params.body)
        self.assertIs(None, params.orig_body)
        self.assertEqual('jrandom@example.com', params.to)
        self.assertEqual('[MERGE] This code rox', params.subject)
        self.assertEqual(directive, params.directive)
        self.assertEqual('foo-1', params.basename)

    def test_body_hook_chaining(self):
        test_hook1 = TestHook('foo')
        test_hook2 = TestHook('bar')
        client = self.compose_with_hooks([test_hook1, test_hook2])[0]
        self.assertEqual(None, test_hook1.calls[0].body)
        self.assertEqual(None, test_hook1.calls[0].orig_body)
        self.assertEqual('foo', test_hook2.calls[0].body)
        self.assertEqual(None, test_hook2.calls[0].orig_body)
        self.assertEqual('bar', client.body)



# Copyright (C) 2006-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Weave-era BzrDir formats."""

from io import BytesIO

from ... import errors, lockable_files
from ...bzr.bzrdir import BzrDir, BzrDirFormat, BzrDirMetaFormat1
from ...controldir import (ControlDir, Converter, MustHaveWorkingTree,
                           NoColocatedBranchSupport, format_registry)
from ...i18n import gettext
from ...lazy_import import lazy_import
from ...transport import NoSuchFile, get_transport, local

lazy_import(globals(), """
import os

from breezy import (
    branch as _mod_branch,,
    graph,
    lockdir,
    osutils,
    revision as _mod_revision,
    trace,
    ui,
    urlutils,
    )
from breezy.bzr import (
    versionedfile,
    weave,
    xml5,
    )
from breezy.plugins.weave_fmt.store.versioned import VersionedFileStore
from breezy.transactions import WriteTransaction
from breezy.plugins.weave_fmt import xml4
""")


class BzrDirFormatAllInOne(BzrDirFormat):
    """Common class for formats before meta-dirs."""

    fixed_components = True

    def initialize_on_transport_ex(self, transport, use_existing_dir=False,
                                   create_prefix=False, force_new_repo=False, stacked_on=None,
                                   stack_on_pwd=None, repo_format_name=None, make_working_trees=None,
                                   shared_repo=False):
        """See ControlDir.initialize_on_transport_ex."""
        require_stacking = (stacked_on is not None)
        # Format 5 cannot stack, but we've been asked to - actually init
        # a Meta1Dir
        if require_stacking:
            format = BzrDirMetaFormat1()
            return format.initialize_on_transport_ex(transport,
                                                     use_existing_dir=use_existing_dir, create_prefix=create_prefix,
                                                     force_new_repo=force_new_repo, stacked_on=stacked_on,
                                                     stack_on_pwd=stack_on_pwd, repo_format_name=repo_format_name,
                                                     make_working_trees=make_working_trees, shared_repo=shared_repo)
        return BzrDirFormat.initialize_on_transport_ex(self, transport,
                                                       use_existing_dir=use_existing_dir, create_prefix=create_prefix,
                                                       force_new_repo=force_new_repo, stacked_on=stacked_on,
                                                       stack_on_pwd=stack_on_pwd, repo_format_name=repo_format_name,
                                                       make_working_trees=make_working_trees, shared_repo=shared_repo)

    @classmethod
    def from_string(cls, format_string):
        if format_string != cls.get_format_string():
            raise AssertionError("unexpected format string %r" % format_string)
        return cls()


class BzrDirFormat5(BzrDirFormatAllInOne):
    """Bzr control format 5.

    This format is a combined format for working tree, branch and repository.
    It has:
     - Format 2 working trees [always]
     - Format 4 branches [always]
     - Format 5 repositories [always]
       Unhashed stores in the repository.
    """

    _lock_class = lockable_files.TransportLock

    def __eq__(self, other):
        return isinstance(self, type(other))

    @classmethod
    def get_format_string(cls):
        """See BzrDirFormat.get_format_string()."""
        return b"Bazaar-NG branch, format 5\n"

    def get_branch_format(self):
        from .branch import BzrBranchFormat4
        return BzrBranchFormat4()

    def get_format_description(self):
        """See ControlDirFormat.get_format_description()."""
        return "All-in-one format 5"

    def get_converter(self, format=None):
        """See ControlDirFormat.get_converter()."""
        # there is one and only one upgrade path here.
        return ConvertBzrDir5To6()

    def _initialize_for_clone(self, url):
        return self.initialize_on_transport(get_transport(url), _cloning=True)

    def initialize_on_transport(self, transport, _cloning=False):
        """Format 5 dirs always have working tree, branch and repository.

        Except when they are being cloned.
        """
        from .branch import BzrBranchFormat4
        from .repository import RepositoryFormat5
        result = (super().initialize_on_transport(transport))
        RepositoryFormat5().initialize(result, _internal=True)
        if not _cloning:
            branch = BzrBranchFormat4().initialize(result)
            result._init_workingtree()
        return result

    def network_name(self):
        return self.get_format_string()

    def _open(self, transport):
        """See BzrDirFormat._open."""
        return BzrDir5(transport, self)

    def __return_repository_format(self):
        """Circular import protection."""
        from .repository import RepositoryFormat5
        return RepositoryFormat5()
    repository_format = property(__return_repository_format)


class BzrDirFormat6(BzrDirFormatAllInOne):
    """Bzr control format 6.

    This format is a combined format for working tree, branch and repository.
    It has:
     - Format 2 working trees [always]
     - Format 4 branches [always]
     - Format 6 repositories [always]
    """

    _lock_class = lockable_files.TransportLock

    def __eq__(self, other):
        return isinstance(self, type(other))

    @classmethod
    def get_format_string(cls):
        """See BzrDirFormat.get_format_string()."""
        return b"Bazaar-NG branch, format 6\n"

    def get_format_description(self):
        """See ControlDirFormat.get_format_description()."""
        return "All-in-one format 6"

    def get_branch_format(self):
        from .branch import BzrBranchFormat4
        return BzrBranchFormat4()

    def get_converter(self, format=None):
        """See ControlDirFormat.get_converter()."""
        # there is one and only one upgrade path here.
        return ConvertBzrDir6ToMeta()

    def _initialize_for_clone(self, url):
        return self.initialize_on_transport(get_transport(url), _cloning=True)

    def initialize_on_transport(self, transport, _cloning=False):
        """Format 6 dirs always have working tree, branch and repository.

        Except when they are being cloned.
        """
        from .branch import BzrBranchFormat4
        from .repository import RepositoryFormat6
        result = super().initialize_on_transport(transport)
        RepositoryFormat6().initialize(result, _internal=True)
        if not _cloning:
            branch = BzrBranchFormat4().initialize(result)
            result._init_workingtree()
        return result

    def network_name(self):
        return self.get_format_string()

    def _open(self, transport):
        """See BzrDirFormat._open."""
        return BzrDir6(transport, self)

    def __return_repository_format(self):
        """Circular import protection."""
        from .repository import RepositoryFormat6
        return RepositoryFormat6()
    repository_format = property(__return_repository_format)


class ConvertBzrDir4To5(Converter):
    """Converts format 4 bzr dirs to format 5."""

    def __init__(self):
        super().__init__()
        self.converted_revs = set()
        self.absent_revisions = set()
        self.text_count = 0
        self.revisions = {}

    def convert(self, to_convert, pb):
        """See Converter.convert()."""
        self.controldir = to_convert
        with ui.ui_factory.nested_progress_bar() as self.pb:
            ui.ui_factory.note(gettext('starting upgrade from format 4 to 5'))
            if isinstance(self.controldir.transport, local.LocalTransport):
                self.controldir.get_workingtree_transport(
                    None).delete('stat-cache')
            self._convert_to_weaves()
            return ControlDir.open(self.controldir.user_url)

    def _convert_to_weaves(self):
        ui.ui_factory.note(gettext(
            'note: upgrade may be faster if all store files are ungzipped first'))
        try:
            # TODO permissions
            stat = self.controldir.transport.stat('weaves')
            if not S_ISDIR(stat.st_mode):
                self.controldir.transport.delete('weaves')
                self.controldir.transport.mkdir('weaves')
        except NoSuchFile:
            self.controldir.transport.mkdir('weaves')
        # deliberately not a WeaveFile as we want to build it up slowly.
        self.inv_weave = weave.Weave('inventory')
        # holds in-memory weaves for all files
        self.text_weaves = {}
        self.controldir.transport.delete('branch-format')
        self.branch = self.controldir.open_branch()
        self._convert_working_inv()
        rev_history = self.branch._revision_history()
        # to_read is a stack holding the revisions we still need to process;
        # appending to it adds new highest-priority revisions
        self.known_revisions = set(rev_history)
        self.to_read = rev_history[-1:]
        while self.to_read:
            rev_id = self.to_read.pop()
            if (rev_id not in self.revisions and
                    rev_id not in self.absent_revisions):
                self._load_one_rev(rev_id)
        self.pb.clear()
        to_import = self._make_order()
        for i, rev_id in enumerate(to_import):
            self.pb.update(gettext('converting revision'), i, len(to_import))
            self._convert_one_rev(rev_id)
        self.pb.clear()
        self._write_all_weaves()
        self._write_all_revs()
        ui.ui_factory.note(gettext('upgraded to weaves:'))
        ui.ui_factory.note('  ' + gettext('%6d revisions and inventories') %
                           len(self.revisions))
        ui.ui_factory.note('  ' + gettext('%6d revisions not present') %
                           len(self.absent_revisions))
        ui.ui_factory.note('  ' + gettext('%6d texts') % self.text_count)
        self._cleanup_spare_files_after_format4()
        self.branch._transport.put_bytes(
            'branch-format',
            BzrDirFormat5().get_format_string(),
            mode=self.controldir._get_file_mode())

    def _cleanup_spare_files_after_format4(self):
        # FIXME working tree upgrade foo.
        for n in 'merged-patches', 'pending-merged-patches':
            try:
                ## assert os.path.getsize(p) == 0
                self.controldir.transport.delete(n)
            except NoSuchFile:
                pass
        self.controldir.transport.delete_tree('inventory-store')
        self.controldir.transport.delete_tree('text-store')

    def _convert_working_inv(self):
        inv = xml4.serializer_v4.read_inventory(
            self.branch._transport.get('inventory'))
        f = BytesIO()
        xml5.serializer_v5.write_inventory(inv, f, working=True)
        self.branch._transport.put_bytes('inventory', f.getvalue(),
                                         mode=self.controldir._get_file_mode())

    def _write_all_weaves(self):
        controlweaves = VersionedFileStore(self.controldir.transport, prefixed=False,
                                           versionedfile_class=weave.WeaveFile)
        weave_transport = self.controldir.transport.clone('weaves')
        weaves = VersionedFileStore(weave_transport, prefixed=False,
                                    versionedfile_class=weave.WeaveFile)
        transaction = WriteTransaction()

        try:
            i = 0
            for file_id, file_weave in self.text_weaves.items():
                self.pb.update(gettext('writing weave'), i,
                               len(self.text_weaves))
                weaves._put_weave(file_id, file_weave, transaction)
                i += 1
            self.pb.update(gettext('inventory'), 0, 1)
            controlweaves._put_weave(b'inventory', self.inv_weave, transaction)
            self.pb.update(gettext('inventory'), 1, 1)
        finally:
            self.pb.clear()

    def _write_all_revs(self):
        """Write all revisions out in new form."""
        self.controldir.transport.delete_tree('revision-store')
        self.controldir.transport.mkdir('revision-store')
        revision_transport = self.controldir.transport.clone('revision-store')
        # TODO permissions
        from ...bzr.xml5 import serializer_v5
        from .repository import RevisionTextStore
        revision_store = RevisionTextStore(revision_transport,
                                           serializer_v5, False, versionedfile.PrefixMapper(),
                                           lambda: True, lambda: True)
        try:
            for i, rev_id in enumerate(self.converted_revs):
                self.pb.update(gettext('write revision'), i,
                               len(self.converted_revs))
                lines = serializer_v5.write_revision_to_lines(
                    self.revisions[rev_id])
                key = (rev_id,)
                revision_store.add_lines(key, None, lines)
        finally:
            self.pb.clear()

    def _load_one_rev(self, rev_id):
        """Load a revision object into memory.

        Any parents not either loaded or abandoned get queued to be
        loaded."""
        self.pb.update(gettext('loading revision'),
                       len(self.revisions),
                       len(self.known_revisions))
        if not self.branch.repository.has_revision(rev_id):
            self.pb.clear()
            ui.ui_factory.note(gettext('revision {%s} not present in branch; '
                                       'will be converted as a ghost') %
                               rev_id)
            self.absent_revisions.add(rev_id)
        else:
            rev = self.branch.repository.get_revision(rev_id)
            for parent_id in rev.parent_ids:
                self.known_revisions.add(parent_id)
                self.to_read.append(parent_id)
            self.revisions[rev_id] = rev

    def _load_old_inventory(self, rev_id):
        with self.branch.repository.inventory_store.get(rev_id) as f:
            inv = xml4.serializer_v4.read_inventory(f)
        inv.revision_id = rev_id
        rev = self.revisions[rev_id]
        return inv

    def _load_updated_inventory(self, rev_id):
        inv_xml = self.inv_weave.get_lines(rev_id)
        inv = xml5.serializer_v5.read_inventory_from_lines(inv_xml, rev_id)
        return inv

    def _convert_one_rev(self, rev_id):
        """Convert revision and all referenced objects to new format."""
        rev = self.revisions[rev_id]
        inv = self._load_old_inventory(rev_id)
        present_parents = [p for p in rev.parent_ids
                           if p not in self.absent_revisions]
        self._convert_revision_contents(rev, inv, present_parents)
        self._store_new_inv(rev, inv, present_parents)
        self.converted_revs.add(rev_id)

    def _store_new_inv(self, rev, inv, present_parents):
        new_inv_xml = xml5.serializer_v5.write_inventory_to_lines(inv)
        new_inv_sha1 = osutils.sha_strings(new_inv_xml)
        self.inv_weave.add_lines(rev.revision_id,
                                 present_parents,
                                 new_inv_xml)
        rev.inventory_sha1 = new_inv_sha1

    def _convert_revision_contents(self, rev, inv, present_parents):
        """Convert all the files within a revision.

        Also upgrade the inventory to refer to the text revision ids."""
        rev_id = rev.revision_id
        trace.mutter('converting texts of revision {%s}', rev_id)
        parent_invs = list(map(self._load_updated_inventory, present_parents))
        entries = inv.iter_entries()
        next(entries)
        for path, ie in entries:
            self._convert_file_version(rev, ie, parent_invs)

    def _convert_file_version(self, rev, ie, parent_invs):
        """Convert one version of one file.

        The file needs to be added into the weave if it is a merge
        of >=2 parents or if it's changed from its parent.
        """
        file_id = ie.file_id
        rev_id = rev.revision_id
        w = self.text_weaves.get(file_id)
        if w is None:
            w = weave.Weave(file_id)
            self.text_weaves[file_id] = w
        text_changed = False
        parent_candiate_entries = ie.parent_candidates(parent_invs)
        heads = graph.Graph(self).heads(parent_candiate_entries)
        # XXX: Note that this is unordered - and this is tolerable because
        # the previous code was also unordered.
        previous_entries = {head: parent_candiate_entries[head]
                            for head in heads}
        self.snapshot_ie(previous_entries, ie, w, rev_id)

    def get_parent_map(self, revision_ids):
        """See graph.StackedParentsProvider.get_parent_map"""
        return {revision_id: self.revisions[revision_id]
                    for revision_id in revision_ids
                    if revision_id in self.revisions}

    def snapshot_ie(self, previous_revisions, ie, w, rev_id):
        # TODO: convert this logic, which is ~= snapshot to
        # a call to:. This needs the path figured out. rather than a work_tree
        # a v4 revision_tree can be given, or something that looks enough like
        # one to give the file content to the entry if it needs it.
        # and we need something that looks like a weave store for snapshot to
        # save against.
        #ie.snapshot(rev, PATH, previous_revisions, REVISION_TREE, InMemoryWeaveStore(self.text_weaves))
        if len(previous_revisions) == 1:
            previous_ie = next(iter(previous_revisions.values()))
            if ie._unchanged(previous_ie):
                ie.revision = previous_ie.revision
                return
        if ie.has_text():
            with self.branch.repository._text_store.get(ie.text_id) as f:
                file_lines = f.readlines()
            w.add_lines(rev_id, previous_revisions, file_lines)
            self.text_count += 1
        else:
            w.add_lines(rev_id, previous_revisions, [])
        ie.revision = rev_id

    def _make_order(self):
        """Return a suitable order for importing revisions.

        The order must be such that an revision is imported after all
        its (present) parents.
        """
        todo = set(self.revisions)
        done = self.absent_revisions.copy()
        order = []
        while todo:
            # scan through looking for a revision whose parents
            # are all done
            for rev_id in sorted(list(todo)):
                rev = self.revisions[rev_id]
                parent_ids = set(rev.parent_ids)
                if parent_ids.issubset(done):
                    # can take this one now
                    order.append(rev_id)
                    todo.remove(rev_id)
                    done.add(rev_id)
        return order


class ConvertBzrDir5To6(Converter):
    """Converts format 5 bzr dirs to format 6."""

    def convert(self, to_convert, pb):
        """See Converter.convert()."""
        self.controldir = to_convert
        with ui.ui_factory.nested_progress_bar() as pb:
            ui.ui_factory.note(gettext('starting upgrade from format 5 to 6'))
            self._convert_to_prefixed()
            return ControlDir.open(self.controldir.user_url)

    def _convert_to_prefixed(self):
        from .store import TransportStore
        self.controldir.transport.delete('branch-format')
        for store_name in ["weaves", "revision-store"]:
            ui.ui_factory.note(gettext("adding prefixes to %s") % store_name)
            store_transport = self.controldir.transport.clone(store_name)
            store = TransportStore(store_transport, prefixed=True)
            for urlfilename in store_transport.list_dir('.'):
                filename = urlutils.unescape(urlfilename)
                if (filename.endswith(".weave")
                    or filename.endswith(".gz")
                        or filename.endswith(".sig")):
                    file_id, suffix = os.path.splitext(filename)
                else:
                    file_id = filename
                    suffix = ''
                new_name = store._mapper.map(
                    (file_id.encode('utf-8'),)) + suffix
                # FIXME keep track of the dirs made RBC 20060121
                try:
                    store_transport.move(filename, new_name)
                except NoSuchFile:  # catches missing dirs strangely enough
                    store_transport.mkdir(osutils.dirname(new_name))
                    store_transport.move(filename, new_name)
        self.controldir.transport.put_bytes(
            'branch-format',
            BzrDirFormat6().get_format_string(),
            mode=self.controldir._get_file_mode())


class ConvertBzrDir6ToMeta(Converter):
    """Converts format 6 bzr dirs to metadirs."""

    def convert(self, to_convert, pb):
        """See Converter.convert()."""
        from ...bzr.fullhistory import BzrBranchFormat5
        from .repository import RepositoryFormat7
        self.controldir = to_convert
        self.pb = ui.ui_factory.nested_progress_bar()
        self.count = 0
        self.total = 20  # the steps we know about
        self.garbage_inventories = []
        self.dir_mode = self.controldir._get_dir_mode()
        self.file_mode = self.controldir._get_file_mode()

        ui.ui_factory.note(
            gettext('starting upgrade from format 6 to metadir'))
        self.controldir.transport.put_bytes(
            'branch-format',
            b"Converting to format 6",
            mode=self.file_mode)
        # its faster to move specific files around than to open and use the apis...
        # first off, nuke ancestry.weave, it was never used.
        try:
            self.step(gettext('Removing ancestry.weave'))
            self.controldir.transport.delete('ancestry.weave')
        except NoSuchFile:
            pass
        # find out whats there
        self.step(gettext('Finding branch files'))
        last_revision = self.controldir.open_branch().last_revision()
        bzrcontents = self.controldir.transport.list_dir('.')
        for name in bzrcontents:
            if name.startswith('basis-inventory.'):
                self.garbage_inventories.append(name)
        # create new directories for repository, working tree and branch
        repository_names = [('inventory.weave', True),
                            ('revision-store', True),
                            ('weaves', True)]
        self.step(gettext('Upgrading repository') + '  ')
        self.controldir.transport.mkdir('repository', mode=self.dir_mode)
        self.make_lock('repository')
        # we hard code the formats here because we are converting into
        # the meta format. The meta format upgrader can take this to a
        # future format within each component.
        self.put_format('repository', RepositoryFormat7())
        for entry in repository_names:
            self.move_entry('repository', entry)

        self.step(gettext('Upgrading branch') + '      ')
        self.controldir.transport.mkdir('branch', mode=self.dir_mode)
        self.make_lock('branch')
        self.put_format('branch', BzrBranchFormat5())
        branch_files = [('revision-history', True),
                        ('branch-name', True),
                        ('parent', False)]
        for entry in branch_files:
            self.move_entry('branch', entry)

        checkout_files = [('pending-merges', True),
                          ('inventory', True),
                          ('stat-cache', False)]
        # If a mandatory checkout file is not present, the branch does not have
        # a functional checkout. Do not create a checkout in the converted
        # branch.
        for name, mandatory in checkout_files:
            if mandatory and name not in bzrcontents:
                has_checkout = False
                break
        else:
            has_checkout = True
        if not has_checkout:
            ui.ui_factory.note(gettext('No working tree.'))
            # If some checkout files are there, we may as well get rid of them.
            for name, mandatory in checkout_files:
                if name in bzrcontents:
                    self.controldir.transport.delete(name)
        else:
            from ...bzr.workingtree_3 import WorkingTreeFormat3
            self.step(gettext('Upgrading working tree'))
            self.controldir.transport.mkdir('checkout', mode=self.dir_mode)
            self.make_lock('checkout')
            self.put_format('checkout', WorkingTreeFormat3())
            for path in self.garbage_inventories:
                self.controldir.transport.delete(path)
            for entry in checkout_files:
                self.move_entry('checkout', entry)
            if last_revision is not None:
                self.controldir.transport.put_bytes(
                    'checkout/last-revision', last_revision)
        self.controldir.transport.put_bytes(
            'branch-format',
            BzrDirMetaFormat1().get_format_string(),
            mode=self.file_mode)
        self.pb.finished()
        return ControlDir.open(self.controldir.user_url)

    def make_lock(self, name):
        """Make a lock for the new control dir name."""
        self.step(gettext('Make %s lock') % name)
        ld = lockdir.LockDir(self.controldir.transport,
                             '%s/lock' % name,
                             file_modebits=self.file_mode,
                             dir_modebits=self.dir_mode)
        ld.create()

    def move_entry(self, new_dir, entry):
        """Move then entry name into new_dir."""
        name = entry[0]
        mandatory = entry[1]
        self.step(gettext('Moving %s') % name)
        try:
            self.controldir.transport.move(name, '{}/{}'.format(new_dir, name))
        except NoSuchFile:
            if mandatory:
                raise

    def put_format(self, dirname, format):
        self.controldir.transport.put_bytes('%s/format' % dirname,
                                            format.get_format_string(),
                                            self.file_mode)


class BzrDirFormat4(BzrDirFormat):
    """Bzr dir format 4.

    This format is a combined format for working tree, branch and repository.
    It has:
     - Format 1 working trees [always]
     - Format 4 branches [always]
     - Format 4 repositories [always]

    This format is deprecated: it indexes texts using a text it which is
    removed in format 5; write support for this format has been removed.
    """

    _lock_class = lockable_files.TransportLock

    def __eq__(self, other):
        return isinstance(self, type(other))

    @classmethod
    def get_format_string(cls):
        """See BzrDirFormat.get_format_string()."""
        return b"Bazaar-NG branch, format 0.0.4\n"

    def get_format_description(self):
        """See ControlDirFormat.get_format_description()."""
        return "All-in-one format 4"

    def get_converter(self, format=None):
        """See ControlDirFormat.get_converter()."""
        # there is one and only one upgrade path here.
        return ConvertBzrDir4To5()

    def initialize_on_transport(self, transport):
        """Format 4 branches cannot be created."""
        raise errors.UninitializableFormat(self)

    def is_supported(self):
        """Format 4 is not supported.

        It is not supported because the model changed from 4 to 5 and the
        conversion logic is expensive - so doing it on the fly was not
        feasible.
        """
        return False

    def network_name(self):
        return self.get_format_string()

    def _open(self, transport):
        """See BzrDirFormat._open."""
        return BzrDir4(transport, self)

    def __return_repository_format(self):
        """Circular import protection."""
        from .repository import RepositoryFormat4
        return RepositoryFormat4()
    repository_format = property(__return_repository_format)

    @classmethod
    def from_string(cls, format_string):
        if format_string != cls.get_format_string():
            raise AssertionError("unexpected format string %r" % format_string)
        return cls()


class BzrDirPreSplitOut(BzrDir):
    """A common class for the all-in-one formats."""

    def __init__(self, _transport, _format):
        """See ControlDir.__init__."""
        super().__init__(_transport, _format)
        self._control_files = lockable_files.LockableFiles(
            self.get_branch_transport(None),
            self._format._lock_file_name,
            self._format._lock_class)

    def break_lock(self):
        """Pre-splitout bzrdirs do not suffer from stale locks."""
        raise NotImplementedError(self.break_lock)

    def cloning_metadir(self, require_stacking=False):
        """Produce a metadir suitable for cloning with."""
        if require_stacking:
            return format_registry.make_controldir('1.6')
        return self._format.__class__()

    def clone(self, url, revision_id=None, force_new_repo=False,
              preserve_stacking=False, tag_selector=None):
        """See ControlDir.clone().

        force_new_repo has no effect, since this family of formats always
        require a new repository.
        preserve_stacking has no effect, since no source branch using this
        family of formats can be stacked, so there is no stacking to preserve.
        """
        self._make_tail(url)
        result = self._format._initialize_for_clone(url)
        self.open_repository().clone(result, revision_id=revision_id)
        from_branch = self.open_branch()
        from_branch.clone(result, revision_id=revision_id, tag_selector=tag_selector)
        try:
            tree = self.open_workingtree()
        except errors.NotLocalUrl:
            # make a new one, this format always has to have one.
            result._init_workingtree()
        else:
            tree.clone(result)
        return result

    def create_branch(self, name=None, repository=None,
                      append_revisions_only=None):
        """See ControlDir.create_branch."""
        if repository is not None:
            raise NotImplementedError(
                "create_branch(repository=<not None>) on {!r}".format(self))
        return self._format.get_branch_format().initialize(self, name=name,
                                                           append_revisions_only=append_revisions_only)

    def destroy_branch(self, name=None):
        """See ControlDir.destroy_branch."""
        raise errors.UnsupportedOperation(self.destroy_branch, self)

    def create_repository(self, shared=False):
        """See ControlDir.create_repository."""
        if shared:
            raise errors.IncompatibleFormat('shared repository', self._format)
        return self.open_repository()

    def destroy_repository(self):
        """See ControlDir.destroy_repository."""
        raise errors.UnsupportedOperation(self.destroy_repository, self)

    def create_workingtree(self, revision_id=None, from_branch=None,
                           accelerator_tree=None, hardlink=False):
        """See ControlDir.create_workingtree."""
        # The workingtree is sometimes created when the bzrdir is created,
        # but not when cloning.

        # this looks buggy but is not -really-
        # because this format creates the workingtree when the bzrdir is
        # created
        # clone and sprout will have set the revision_id
        # and that will have set it for us, its only
        # specific uses of create_workingtree in isolation
        # that can do wonky stuff here, and that only
        # happens for creating checkouts, which cannot be
        # done on this format anyway. So - acceptable wart.
        if hardlink:
            warning("can't support hardlinked working trees in %r"
                    % (self,))
        try:
            result = self.open_workingtree(recommend_upgrade=False)
        except NoSuchFile:
            result = self._init_workingtree()
        if revision_id is not None:
            if revision_id == _mod_revision.NULL_REVISION:
                result.set_parent_ids([])
            else:
                result.set_parent_ids([revision_id])
        return result

    def _init_workingtree(self):
        from .workingtree import WorkingTreeFormat2
        try:
            return WorkingTreeFormat2().initialize(self)
        except errors.NotLocalUrl:
            # Even though we can't access the working tree, we need to
            # create its control files.
            return WorkingTreeFormat2()._stub_initialize_on_transport(
                self.transport, self._control_files._file_mode)

    def destroy_workingtree(self):
        """See ControlDir.destroy_workingtree."""
        raise errors.UnsupportedOperation(self.destroy_workingtree, self)

    def destroy_workingtree_metadata(self):
        """See ControlDir.destroy_workingtree_metadata."""
        raise errors.UnsupportedOperation(self.destroy_workingtree_metadata,
                                          self)

    def get_branch_transport(self, branch_format, name=None):
        """See BzrDir.get_branch_transport()."""
        if name:
            raise NoColocatedBranchSupport(self)
        if branch_format is None:
            return self.transport
        try:
            branch_format.get_format_string()
        except NotImplementedError:
            return self.transport
        raise errors.IncompatibleFormat(branch_format, self._format)

    def get_repository_transport(self, repository_format):
        """See BzrDir.get_repository_transport()."""
        if repository_format is None:
            return self.transport
        try:
            repository_format.get_format_string()
        except NotImplementedError:
            return self.transport
        raise errors.IncompatibleFormat(repository_format, self._format)

    def get_workingtree_transport(self, workingtree_format):
        """See BzrDir.get_workingtree_transport()."""
        if workingtree_format is None:
            return self.transport
        try:
            workingtree_format.get_format_string()
        except NotImplementedError:
            return self.transport
        raise errors.IncompatibleFormat(workingtree_format, self._format)

    def needs_format_conversion(self, format):
        """See ControlDir.needs_format_conversion()."""
        # if the format is not the same as the system default,
        # an upgrade is needed.
        return not isinstance(self._format, format.__class__)

    def open_branch(self, name=None, unsupported=False,
                    ignore_fallbacks=False, possible_transports=None):
        """See ControlDir.open_branch."""
        from .branch import BzrBranchFormat4
        format = BzrBranchFormat4()
        format.check_support_status(unsupported)
        return format.open(self, name, _found=True,
                           possible_transports=possible_transports)

    def sprout(self, url, revision_id=None, force_new_repo=False,
               recurse=None, possible_transports=None, accelerator_tree=None,
               hardlink=False, stacked=False, create_tree_if_local=True,
               source_branch=None):
        """See ControlDir.sprout()."""
        if source_branch is not None:
            my_branch = self.open_branch()
            if source_branch.base != my_branch.base:
                raise AssertionError(
                    "source branch %r is not within %r with branch %r" %
                    (source_branch, self, my_branch))
        if stacked:
            raise _mod_branch.UnstackableBranchFormat(
                self._format, self.root_transport.base)
        if not create_tree_if_local:
            raise MustHaveWorkingTree(
                self._format, self.root_transport.base)
        from .workingtree import WorkingTreeFormat2
        self._make_tail(url)
        result = self._format._initialize_for_clone(url)
        try:
            self.open_repository().clone(result, revision_id=revision_id)
        except errors.NoRepositoryPresent:
            pass
        try:
            self.open_branch().sprout(result, revision_id=revision_id)
        except errors.NotBranchError:
            pass

        # we always want a working tree
        WorkingTreeFormat2().initialize(result,
                                        accelerator_tree=accelerator_tree,
                                        hardlink=hardlink)
        return result

    def set_branch_reference(self, target_branch, name=None):
        from ...bzr.branch import BranchReferenceFormat
        if name is not None:
            raise NoColocatedBranchSupport(self)
        raise errors.IncompatibleFormat(BranchReferenceFormat, self._format)


class BzrDir4(BzrDirPreSplitOut):
    """A .bzr version 4 control object.

    This is a deprecated format and may be removed after sept 2006.
    """

    def create_repository(self, shared=False):
        """See ControlDir.create_repository."""
        return self._format.repository_format.initialize(self, shared)

    def needs_format_conversion(self, format):
        """Format 4 dirs are always in need of conversion."""
        return True

    def open_repository(self):
        """See ControlDir.open_repository."""
        from .repository import RepositoryFormat4
        return RepositoryFormat4().open(self, _found=True)


class BzrDir5(BzrDirPreSplitOut):
    """A .bzr version 5 control object.

    This is a deprecated format and may be removed after sept 2006.
    """

    def has_workingtree(self):
        """See ControlDir.has_workingtree."""
        return True

    def open_repository(self):
        """See ControlDir.open_repository."""
        from .repository import RepositoryFormat5
        return RepositoryFormat5().open(self, _found=True)

    def open_workingtree(self, unsupported=False,
                         recommend_upgrade=True):
        """See ControlDir.create_workingtree."""
        from .workingtree import WorkingTreeFormat2
        wt_format = WorkingTreeFormat2()
        # we don't warn here about upgrades; that ought to be handled for the
        # bzrdir as a whole
        return wt_format.open(self, _found=True)


class BzrDir6(BzrDirPreSplitOut):
    """A .bzr version 6 control object.

    This is a deprecated format and may be removed after sept 2006.
    """

    def has_workingtree(self):
        """See ControlDir.has_workingtree."""
        return True

    def open_repository(self):
        """See ControlDir.open_repository."""
        from .repository import RepositoryFormat6
        return RepositoryFormat6().open(self, _found=True)

    def open_workingtree(self, unsupported=False, recommend_upgrade=True):
        """See ControlDir.create_workingtree."""
        # we don't warn here about upgrades; that ought to be handled for the
        # bzrdir as a whole
        from .workingtree import WorkingTreeFormat2
        return WorkingTreeFormat2().open(self, _found=True)



# Copyright (C) 2010-2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for bzr-git's object store."""

import os
import shutil
import stat

from dulwich.objects import Blob, Tree

from ...branchbuilder import BranchBuilder
from ...bzr.inventory import InventoryDirectory, InventoryFile
from ...errors import NoSuchRevision
from ...graph import DictParentsProvider, Graph
from ...tests import TestCase, TestCaseWithTransport
from ...tests.features import SymlinkFeature
from ..cache import DictGitShaMap
from ..object_store import (BazaarObjectStore, LRUTreeCache,
                            _check_expected_sha, _find_missing_bzr_revids,
                            _tree_to_objects, directory_to_tree)


class ExpectedShaTests(TestCase):

    def setUp(self):
        super().setUp()
        self.obj = Blob()
        self.obj.data = b"foo"

    def test_none(self):
        _check_expected_sha(None, self.obj)

    def test_hex(self):
        _check_expected_sha(
            self.obj.sha().hexdigest().encode('ascii'), self.obj)
        self.assertRaises(AssertionError, _check_expected_sha,
                          b"0" * 40, self.obj)

    def test_binary(self):
        _check_expected_sha(self.obj.sha().digest(), self.obj)
        self.assertRaises(AssertionError, _check_expected_sha,
                          b"x" * 20, self.obj)


class FindMissingBzrRevidsTests(TestCase):

    def _find_missing(self, ancestry, want, have):
        return _find_missing_bzr_revids(
            Graph(DictParentsProvider(ancestry)),
            set(want), set(have))

    def test_simple(self):
        self.assertEqual(set(), self._find_missing({}, [], []))

    def test_up_to_date(self):
        self.assertEqual(set(),
                         self._find_missing({"a": ["b"]}, ["a"], ["a"]))

    def test_one_missing(self):
        self.assertEqual({"a"},
                         self._find_missing({"a": ["b"]}, ["a"], ["b"]))

    def test_two_missing(self):
        self.assertEqual({"a", "b"},
                         self._find_missing({"a": ["b"], "b": ["c"]}, ["a"], ["c"]))

    def test_two_missing_history(self):
        self.assertEqual({"a", "b"},
                         self._find_missing({"a": ["b"], "b": ["c"], "c": ["d"]},
                                            ["a"], ["c"]))


class LRUTreeCacheTests(TestCaseWithTransport):

    def setUp(self):
        super().setUp()
        self.branch = self.make_branch(".")
        self.branch.lock_write()
        self.addCleanup(self.branch.unlock)
        self.cache = LRUTreeCache(self.branch.repository)

    def test_get_not_present(self):
        self.assertRaises(NoSuchRevision, self.cache.revision_tree,
                          "unknown")

    def test_revision_trees(self):
        self.assertRaises(NoSuchRevision, self.cache.revision_trees,
                          ["unknown", "la"])

    def test_iter_revision_trees(self):
        self.assertRaises(NoSuchRevision, self.cache.iter_revision_trees,
                          ["unknown", "la"])

    def test_get(self):
        bb = BranchBuilder(branch=self.branch)
        bb.start_series()
        revid = bb.build_snapshot(None,
                                  [('add', ('', None, 'directory', None)),
                                   ('add', ('foo', b'foo-id',
                                            'file', b'a\nb\nc\nd\ne\n')),
                                   ])
        bb.finish_series()
        tree = self.cache.revision_tree(revid)
        self.assertEqual(revid, tree.get_revision_id())


class BazaarObjectStoreTests(TestCaseWithTransport):

    def setUp(self):
        super().setUp()
        self.branch = self.make_branch(".")
        self.store = BazaarObjectStore(self.branch.repository)

    def test_get_blob(self):
        self.branch.lock_write()
        self.addCleanup(self.branch.unlock)
        b = Blob()
        b.data = b'a\nb\nc\nd\ne\n'
        self.store.lock_read()
        self.addCleanup(self.store.unlock)
        self.assertRaises(KeyError, self.store.__getitem__, b.id)
        bb = BranchBuilder(branch=self.branch)
        bb.start_series()
        bb.build_snapshot(None,
                          [('add', ('', None, 'directory', None)),
                           ('add', ('foo', b'foo-id', 'file', b'a\nb\nc\nd\ne\n')),
                           ])
        bb.finish_series()
        # read locks cache
        self.assertRaises(KeyError, self.store.__getitem__, b.id)
        self.store.unlock()
        self.store.lock_read()
        self.assertEqual(b, self.store[b.id])

    def test_directory_converted_to_symlink(self):
        self.requireFeature(SymlinkFeature(self.test_dir))
        b = Blob()
        b.data = b'trgt'
        self.store.lock_read()
        self.addCleanup(self.store.unlock)
        self.assertRaises(KeyError, self.store.__getitem__, b.id)
        tree = self.branch.controldir.create_workingtree()
        self.build_tree_contents([
            ('foo/', ),
            ('foo/bar', b'a\nb\nc\nd\ne\n')])
        tree.add(['foo', 'foo/bar'])
        revid1 = tree.commit('commit 1')
        shutil.rmtree('foo')
        os.symlink('trgt', 'foo')
        revid2 = tree.commit('commit 2')
        # read locks cache
        self.assertRaises(KeyError, self.store.__getitem__, b.id)
        self.store.unlock()
        self.store.lock_read()
        self.assertEqual(b, self.store[b.id])

    def test_get_raw(self):
        self.branch.lock_write()
        self.addCleanup(self.branch.unlock)
        b = Blob()
        b.data = b'a\nb\nc\nd\ne\n'
        self.store.lock_read()
        self.addCleanup(self.store.unlock)
        self.assertRaises(KeyError, self.store.get_raw, b.id)
        bb = BranchBuilder(branch=self.branch)
        bb.start_series()
        bb.build_snapshot(None,
                          [('add', ('', None, 'directory', None)),
                           ('add', ('foo', b'foo-id', 'file', b'a\nb\nc\nd\ne\n')),
                           ])
        bb.finish_series()
        # read locks cache
        self.assertRaises(KeyError, self.store.get_raw, b.id)
        self.store.unlock()
        self.store.lock_read()
        self.assertEqual(b.as_raw_string(), self.store.get_raw(b.id)[1])

    def test_contains(self):
        self.branch.lock_write()
        self.addCleanup(self.branch.unlock)
        b = Blob()
        b.data = b'a\nb\nc\nd\ne\n'
        self.store.lock_read()
        self.addCleanup(self.store.unlock)
        self.assertFalse(b.id in self.store)
        bb = BranchBuilder(branch=self.branch)
        bb.start_series()
        bb.build_snapshot(None,
                          [('add', ('', None, 'directory', None)),
                           ('add', ('foo', b'foo-id', 'file', b'a\nb\nc\nd\ne\n')),
                           ])
        bb.finish_series()
        # read locks cache
        self.assertFalse(b.id in self.store)
        self.store.unlock()
        self.store.lock_read()
        self.assertTrue(b.id in self.store)


class TreeToObjectsTests(TestCaseWithTransport):

    def setUp(self):
        super().setUp()
        self.idmap = DictGitShaMap()

    def test_no_changes(self):
        tree = self.make_branch_and_tree('.')
        self.addCleanup(tree.lock_read().unlock)
        entries = list(_tree_to_objects(tree, [tree], self.idmap, {}))
        self.assertEqual([], entries)

    def test_with_gitdir(self):
        tree = self.make_branch_and_tree('.')
        self.build_tree(['.git', 'foo'])
        tree.add(['.git', 'foo'])
        revid = tree.commit('commit')
        revtree = tree.branch.repository.revision_tree(revid)
        self.addCleanup(revtree.lock_read().unlock)
        entries = list(_tree_to_objects(revtree, [], self.idmap, {}))
        self.assertEqual(['foo', ''], [p[0] for p in entries])

    def test_merge(self):
        basis_tree = self.make_branch_and_tree('base')
        self.build_tree(['base/foo/'])
        basis_tree.add(['foo'])
        basis_rev = basis_tree.commit('foo')
        basis_revtree = basis_tree.branch.repository.revision_tree(basis_rev)

        tree_a = self.make_branch_and_tree('a')
        tree_a.pull(basis_tree.branch)

        self.build_tree(['a/foo/file1'])
        self.build_tree(['a/foo/subdir-a/'])
        os.symlink('.', 'a/foo/subdir-a/symlink')
        tree_a.add(['foo/subdir-a', 'foo/subdir-a/symlink'])

        tree_a.add(['foo/file1'])
        rev_a = tree_a.commit('commit a')
        revtree_a = tree_a.branch.repository.revision_tree(rev_a)

        with revtree_a.lock_read():
            entries = list(_tree_to_objects(revtree_a, [basis_revtree],
                           self.idmap, {}))
            objects = {path: obj for (path, obj, key) in entries}
            subdir_a = objects['foo/subdir-a']

        tree_b = self.make_branch_and_tree('b')
        tree_b.pull(basis_tree.branch)
        self.build_tree(['b/foo/subdir/'])
        os.symlink('.', 'b/foo/subdir/symlink')
        tree_b.add(['foo/subdir', 'foo/subdir/symlink'])
        rev_b = tree_b.commit('commit b')
        revtree_b = tree_b.branch.repository.revision_tree(rev_b)
        self.addCleanup(revtree_b.lock_read().unlock)

        with revtree_b.lock_read():
            list(_tree_to_objects(revtree_b, [basis_revtree], self.idmap, {}))

        with tree_a.lock_write():
            tree_a.merge_from_branch(tree_b.branch)
        rev_merge = tree_a.commit('merge')

        revtree_merge = tree_a.branch.basis_tree()
        self.addCleanup(revtree_merge.lock_read().unlock)

        entries = list(_tree_to_objects(
            revtree_merge,
            [tree_a.branch.repository.revision_tree(r)
             for r in revtree_merge.get_parent_ids()],
            self.idmap, {}))
        objects = {path: obj for (path, obj, key) in entries}
        self.assertEqual({'', 'foo', 'foo/subdir'}, set(objects))
        self.assertEqual(
            (stat.S_IFDIR, subdir_a.id), objects['foo'][b'subdir-a'])


class DirectoryToTreeTests(TestCase):

    def test_empty(self):
        t = directory_to_tree('', [], None, {}, None, allow_empty=False)
        self.assertEqual(None, t)

    def test_empty_dir(self):
        child_ie = InventoryDirectory(b'bar', 'bar', b'bar')
        t = directory_to_tree('', [child_ie], lambda p, x: None, {}, None,
                              allow_empty=False)
        self.assertEqual(None, t)

    def test_empty_dir_dummy_files(self):
        child_ie = InventoryDirectory(b'bar', 'bar', b'bar')
        t = directory_to_tree('', [child_ie], lambda p, x: None, {}, ".mydummy",
                              allow_empty=False)
        self.assertTrue(".mydummy" in t)

    def test_empty_root(self):
        child_ie = InventoryDirectory(b'bar', 'bar', b'bar')
        t = directory_to_tree('', [child_ie], lambda p, x: None, {}, None,
                              allow_empty=True)
        self.assertEqual(Tree(), t)

    def test_with_file(self):
        child_ie = InventoryFile(b'bar', 'bar', b'bar')
        b = Blob.from_string(b"bla")
        t1 = directory_to_tree('', [child_ie], lambda p, x: b.id, {}, None,
                               allow_empty=False)
        t2 = Tree()
        t2.add(b"bar", 0o100644, b.id)
        self.assertEqual(t1, t2)

    def test_with_gitdir(self):
        child_ie = InventoryFile(b'bar', 'bar', b'bar')
        git_file_ie = InventoryFile(b'gitid', '.git', b'bar')
        b = Blob.from_string(b"bla")
        t1 = directory_to_tree('', [child_ie, git_file_ie],
                               lambda p, x: b.id, {}, None,
                               allow_empty=False)
        t2 = Tree()
        t2.add(b"bar", 0o100644, b.id)
        self.assertEqual(t1, t2)



# Copyright (C) 2006-2012 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""External tests of 'brz ls'"""

from breezy import ignores, tests


class TestLS(tests.TestCaseWithTransport):

    def setUp(self):
        super().setUp()

        # Create a simple branch that can be used in testing
        ignores._set_user_ignores(['user-ignore'])

        self.wt = self.make_branch_and_tree('.')
        self.build_tree_contents([
                                 ('.bzrignore', b'*.pyo\n'),
                                 ('a', b'hello\n'),
                                 ])

    def ls_equals(self, value, args=None, recursive=True, working_dir=None):
        command = 'ls'
        if args is not None:
            command += ' ' + args
        if recursive:
            command += ' -R'
        out, err = self.run_bzr(command, working_dir=working_dir)
        self.assertEqual('', err)
        self.assertEqualDiff(value, out)

    def test_ls_null_verbose(self):
        # Can't supply both
        self.run_bzr_error(['Cannot set both --verbose and --null'],
                           'ls --verbose --null')

    def test_ls_basic(self):
        """Test the abilities of 'brz ls'"""
        self.ls_equals('.bzrignore\na\n')
        self.ls_equals('.bzrignore\na\n', './')
        self.ls_equals('?        .bzrignore\n'
                       '?        a\n',
                       '--verbose')
        self.ls_equals('.bzrignore\n'
                       'a\n',
                       '--unknown')
        self.ls_equals('', '--ignored')
        self.ls_equals('', '--versioned')
        self.ls_equals('', '-V')
        self.ls_equals('.bzrignore\n'
                       'a\n',
                       '--unknown --ignored --versioned')
        self.ls_equals('.bzrignore\n'
                       'a\n',
                       '--unknown --ignored -V')
        self.ls_equals('', '--ignored --versioned')
        self.ls_equals('', '--ignored -V')
        self.ls_equals('.bzrignore\0a\0', '--null')

    def test_ls_added(self):
        self.wt.add(['a'])
        self.ls_equals('?        .bzrignore\n'
                       'V        a\n',
                       '--verbose')
        self.wt.commit('add')

        self.build_tree(['subdir/'])
        self.ls_equals('?        .bzrignore\n'
                       'V        a\n'
                       '?        subdir/\n', '--verbose')
        self.build_tree(['subdir/b'])
        self.wt.add(['subdir/', 'subdir/b', '.bzrignore'])
        self.ls_equals('V        .bzrignore\n'
                       'V        a\n'
                       'V        subdir/\n'
                       'V        subdir/b\n', '--verbose')

    def test_show_ids(self):
        self.build_tree(['subdir/'])
        self.wt.add(['a', 'subdir'], ids=[b'a-id', b'subdir-id'])
        self.ls_equals(
            '.bzrignore                                         \n'
            'a                                                  a-id\n'
            'subdir/                                            subdir-id\n',
            '--show-ids')
        self.ls_equals(
            '?        .bzrignore\n'
            'V        a                                         a-id\n'
            'V        subdir/                                   subdir-id\n',
            '--show-ids --verbose')
        self.ls_equals('.bzrignore\0\0'
                       'a\0a-id\0'
                       'subdir\0subdir-id\0', '--show-ids --null')

    def test_ls_no_recursive(self):
        self.build_tree(['subdir/', 'subdir/b'])
        self.wt.add(['a', 'subdir/', 'subdir/b', '.bzrignore'])

        self.ls_equals('.bzrignore\n'
                       'a\n'
                       'subdir/\n', recursive=False)

        self.ls_equals('V        .bzrignore\n'
                       'V        a\n'
                       'V        subdir/\n', '--verbose', recursive=False)

        # Check what happens in a sub-directory
        self.ls_equals('b\n', working_dir='subdir')
        self.ls_equals('b\0', '--null', working_dir='subdir')
        self.ls_equals('subdir/b\n', '--from-root', working_dir='subdir')
        self.ls_equals('subdir/b\0', '--from-root --null',
                       working_dir='subdir')
        self.ls_equals('subdir/b\n', '--from-root', recursive=False,
                       working_dir='subdir')

    def test_ls_path(self):
        """If a path is specified, files are listed with that prefix"""
        self.build_tree(['subdir/', 'subdir/b'])
        self.wt.add(['subdir', 'subdir/b'])
        self.ls_equals('subdir/b\n',
                       'subdir')
        self.ls_equals('../.bzrignore\n'
                       '../a\n'
                       '../subdir/\n'
                       '../subdir/b\n',
                       '..', working_dir='subdir')
        self.ls_equals('../.bzrignore\0'
                       '../a\0'
                       '../subdir\0'
                       '../subdir/b\0',
                       '.. --null', working_dir='subdir')
        self.ls_equals('?        ../.bzrignore\n'
                       '?        ../a\n'
                       'V        ../subdir/\n'
                       'V        ../subdir/b\n',
                       '.. --verbose', working_dir='subdir')
        self.run_bzr_error(['cannot specify both --from-root and PATH'],
                           'ls --from-root ..', working_dir='subdir')

    def test_ls_revision(self):
        self.wt.add(['a'])
        self.wt.commit('add')

        self.build_tree(['subdir/'])

        # Check what happens when we supply a specific revision
        self.ls_equals('a\n', '--revision 1')
        self.ls_equals('V        a\n', '--verbose --revision 1')

        self.ls_equals('', '--revision 1', working_dir='subdir')

    def test_ls_branch(self):
        """If a branch is specified, files are listed from it"""
        self.build_tree(['subdir/', 'subdir/b'])
        self.wt.add(['subdir', 'subdir/b'])
        self.wt.commit('committing')
        branch = self.make_branch('branchdir')
        branch.pull(self.wt.branch)
        self.ls_equals('branchdir/subdir/\n'
                       'branchdir/subdir/b\n',
                       'branchdir')
        self.ls_equals('branchdir/subdir/\n'
                       'branchdir/subdir/b\n',
                       'branchdir --revision 1')

    def test_ls_ignored(self):
        # Now try to do ignored files.
        self.wt.add(['a', '.bzrignore'])

        self.build_tree(['blah.py', 'blah.pyo', 'user-ignore'])
        self.ls_equals('.bzrignore\n'
                       'a\n'
                       'blah.py\n'
                       'blah.pyo\n'
                       'user-ignore\n'
                       )
        self.ls_equals('V        .bzrignore\n'
                       'V        a\n'
                       '?        blah.py\n'
                       'I        blah.pyo\n'
                       'I        user-ignore\n', '--verbose')
        self.ls_equals('blah.pyo\n'
                       'user-ignore\n', '--ignored')
        self.ls_equals('blah.py\n', '--unknown')
        self.ls_equals('.bzrignore\n'
                       'a\n', '--versioned')
        self.ls_equals('.bzrignore\n'
                       'a\n', '-V')

    def test_kinds(self):
        self.build_tree(['subdir/'])
        self.ls_equals('.bzrignore\n'
                       'a\n',
                       '--kind=file')
        self.ls_equals('subdir/\n',
                       '--kind=directory')
        self.ls_equals('',
                       '--kind=symlink')
        self.run_bzr_error(['invalid kind specified'], 'ls --kind=pile')

    def test_ls_path_nonrecursive(self):
        self.ls_equals('%s/.bzrignore\n'
                       '%s/a\n'
                       % (self.test_dir, self.test_dir),
                       self.test_dir, recursive=False)

    def test_ls_directory(self):
        """Test --directory option"""
        self.wt = self.make_branch_and_tree('dir')
        self.build_tree(['dir/sub/', 'dir/sub/file'])
        self.wt.add(['sub', 'sub/file'])
        self.wt.commit('commit')
        self.ls_equals('sub/\nsub/file\n', '--directory=dir')
        self.ls_equals('sub/file\n', '-d dir sub')
